# This module is to estimate the sequence or position biases. It provides
# ways to access and save the bias parameters.

import pysam
import numpy as np
import pylab as pl

def norm_pdf(x, mu, sigma):
    return 1 / (sigma*np.sqrt(2*np.pi)) * np.exp(-1/2*((x-mu)/sigma)**2)

class FastaFile:
    """docstring for FastaFile"""
    def __init__(self, fasta_file):
        self.f = pysam.FastaFile(fasta_file)

    def get_seq(self, qref, start, stop):
        """get the sequence in a given region, the start is from 1.
        The start and stop index may still need double check."""
        return self.f.fetch(qref, start-1, stop)

class BiasFile:
    """docstring for BiasFile"""
    def __init__(self, bias_file=None):
        """get the bias parameters from the hdf5 file"""
        self.set_base_chain()
        self.pos5_bias = np.zeros((5, 20))
        self.pos3_bias = np.zeros((5, 20))
        self.pos5_unif = np.zeros((5, 20))
        self.pos3_unif = np.zeros((5, 20))
        self.pos5_prob = np.zeros((5, 20))
        self.pos3_prob = np.zeros((5, 20))
        self.percentile = np.zeros((5, 2))
        self.flen_mean, self.flen_std = 200, 20

        self.seq5_bias, self.seq3_bias = {}, {}
        self.seq5_unif, self.seq3_unif = {}, {}
        self.seq5_prob, self.seq3_prob = {}, {}
        for i in range(len(self.chain_len)):
            self.seq5_bias[str(i)] = np.zeros(4**self.chain_len[i])
            self.seq3_bias[str(i)] = np.zeros(4**self.chain_len[i])
            self.seq5_unif[str(i)] = np.zeros(4**self.chain_len[i])
            self.seq3_unif[str(i)] = np.zeros(4**self.chain_len[i])
            self.seq5_prob[str(i)] = np.zeros(4**self.chain_len[i])
            self.seq3_prob[str(i)] = np.zeros(4**self.chain_len[i])
        
        if bias_file is None: return
        
        fid = open(bias_file, "r")
        all_lines = fid.readlines()
        fid.close()
        
        self.flen_mean = float(all_lines[4].split("\t")[0])
        self.flen_std = float(all_lines[4].split("\t")[1])
        for i in range(5,105):
            a, b = (i-5) // 20, (i-5) % 20
            if b == 0:
                self.percentile[a,:] = all_lines[i].split("|")[0].split("-")
            self.pos5_bias[a,b] = all_lines[i].split("\t")[1]
            self.pos3_bias[a,b] = all_lines[i].split("\t")[2]
            self.pos5_unif[a,b] = all_lines[i].split("\t")[3]
            self.pos3_unif[a,b] = all_lines[i].split("\t")[4]
            self.pos5_prob[a,b] = max(0, self.pos5_bias[a,b] / self.pos5_unif[a,b])
            self.pos3_prob[a,b] = max(0, self.pos3_bias[a,b] / self.pos3_unif[a,b])
        
        ii, cnt = all_lines[105].split("|")[0], -1
        for i in range(105,849):
            if ii == all_lines[i].split("|")[0]: 
                cnt += 1
            else:
                ii = all_lines[i].split("|")[0]
                cnt = 0
            self.seq5_bias[ii][cnt] = all_lines[i].split("\t")[1]
            self.seq3_bias[ii][cnt] = all_lines[i].split("\t")[2]
            self.seq5_unif[ii][cnt] = all_lines[i].split("\t")[3]
            self.seq3_unif[ii][cnt] = all_lines[i].split("\t")[4]
            self.seq5_prob[ii][cnt] = max(0, self.seq5_bias[ii][cnt] / self.seq5_unif[ii][cnt])
            self.seq3_prob[ii][cnt] = max(0, self.seq3_bias[ii][cnt] / self.seq3_unif[ii][cnt])
            self.base_chain[ii][cnt] = all_lines[i].split("\t")[0].split("|")[1]
            
    def set_base_chain(self):
        """set the sub-base chain for the variable-length Markov model (VLMM),
        which was proposed by Reberts et al, Genome Biology, 2011: 
        Figure2 in supp 3. http://genomebiology.com/2011/12/3/r22/"""
        b1 = ["A","T","G","C"]
        b2, b3 = [], []
        for i in b1:
            for j in b1:
                b2.append(j+i)
                for k in b1:
                    b3.append(k+j+i)
        base_comb = [b1, b2, b3]

        self.chain_len = [1]*4 + [2]*3 + [3]*10 + [2]*2 + [1]*2
        self.base_chain = {}
        for i in range(21):
            self.base_chain[str(i)] = base_comb[self.chain_len[i]-1]

    def get_both_bias(seq, loc, ulen, end_num=5):
        """get the bias from the bias parameters"""
        prob = (self.get_seq_bias(seq, end_num) * 
                self.get_pos_bias(loc, ulen, end_num))
        return prob

    def get_seq_bias(self, seq, end_num):
        """get the sequence bias score"""
        if end_num == 5:
            parameters = self.seq5_prob
        elif end_num == 3:
            parameters = self.seq3_prob
        else:
            print("wrong end_num: %s" %str(end_num))
            return None

        prob = 1.0
        for j in range(len(seq)):
            _len = self.chain_len[j]
            _bas = seq[j-_len+1 : j+1]
            _idx = self.base_chain[str(j)].index(_bas)
            prob = prob * parameters[str(j)][_idx]
        return prob

    def get_pos_bias(self, loc, ulen, end_num):
        """get the position bias score, the loc is base pair distance
        from the 5'end of the units"""
        if end_num == 5:
            parameters = self.pos5_prob
        elif end_num == 3:
            parameters = self.pos3_prob
        else:
            print("wrong end_num: %s" %str(end_num))
            return None

        bin1 = (ulen >= self.percentile[:,0]) * (ulen <= self.percentile[:,1])
        bin2 = 20.0 * loc / ulen
        prob = parameters[bin1, bin2]
        return prob

    def set_percentile(self, ulen, K=5):
        """set the percentiles by input the lengths of unitsets, i.e., ulen, 
        and number of percentiles, K."""
        perc_gap = np.linspace(0, 100, K+1)
        _percent = np.percentile(ulen, list(perc_gap))
        self.percentile = np.zeros((K, 2))
        for i in range(K):
            self.percentile[i, 0] = int(_percent[i])+1
            self.percentile[i, 1] = int(_percent[i+1])
            if i == 0:
                self.percentile[i,0] = 0
            elif i==4:
                self.percentile[i,1] = float("inf")

    def set_both_bias(self, seq, loc, ulen, weight, end_num=5, mode="bias"):
        """get the bias from the bias parameters"""
        self.set_seq_bias(seq, weight, end_num, mode)
        self.set_pos_bias(loc, ulen, weight, end_num, mode)

    def set_seq_bias(self, seq, weight, end_num=5, mode="bias"):
        """get the sequence bias score"""
        for j in range(len(seq)):
            _len = self.chain_len[j]
            _bas = seq[j-_len+1 : j+1]
            _idx = self.base_chain[str(j)].index(_bas)
            if end_num == 5:
                if mode == "bias":
                    self.seq5_bias[str(j)][_idx] += weight
                elif mode == "unif":
                    self.seq5_unif[str(j)][_idx] += weight
            else:
                if mode == "bias":
                    self.seq3_bias[str(j)][_idx] += weight
                elif mode == "unif":
                    self.seq3_unif[str(j)][_idx] += weight

    def set_pos_bias(self, loc, ulen, weight, end_num=5, mode="bias"):
        """get the position bias score, the loc is base pair distance
        from the 5'end of the units"""
        bin1 = (ulen >= self.percentile[:,0]) * (ulen <= self.percentile[:,1])
        bin2 = int(20.0 * loc / (ulen + 0.0001))
        if end_num == 5:
            if mode == "bias":
                self.pos5_bias[bin1, bin2] += weight
            elif mode == "unif":
                self.pos5_unif[bin1, bin2] += weight
        else:
            if mode == "bias":
                self.pos3_bias[bin1, bin2] += weight
            elif mode == "unif":
                self.pos3_unif[bin1, bin2] += weight

    def save_file(self, out_file="out_file.bias"):
        """to save the bias file in BIAS FILE FORMAT"""
        fid = open(out_file, "w")
        fid.writelines("# BIAS PARAMETER FORMAT\n")
        fid.writelines("# fragment length: 2 (mean, std), line 5\n")
        fid.writelines("# position bias: 5*20*4 (name, b5, b3, u5, u3), line 6-105\n")
        fid.writelines("# sequence bias: 744*4 (name, b5, b3, u5, u3), line 106-849\n")
        fid.writelines("%.2f\t%.2f\n" %(self.flen_mean, self.flen_std))
        for i in range(self.pos5_bias.shape[0]):
            for j in range(self.pos5_bias.shape[1]):
                aLine = ("%.0f-%.0f|%d\t%.2e\t%.2e\t%.2e\t%.2e\n"
                         %(self.percentile[i,0], self.percentile[i,1], j, self.pos5_bias[i,j], 
                           self.pos3_bias[i,j], self.pos5_unif[i,j], self.pos3_unif[i,j]))
                fid.writelines(aLine)
        for i in sorted(self.base_chain.keys(), key=float):
            for j in range(len(self.base_chain[i])):
                aLine = ("%s|%s\t%.2e\t%.2e\t%.2e\t%.2e\n"
                         %(i, self.base_chain[i][j], self.seq5_bias[i][j], 
                           self.seq3_bias[i][j], self.seq5_unif[i][j], self.seq3_unif[i][j]))
                fid.writelines(aLine)
        fid.close()


    def plot_bias(self, mode=None):
        """plot of bias parameters: flen, pos5, pos3, seq5, seq3"""
        #fragment distribution
        if mode == "flen":
            xx = np.arange(0, 1000)
            yy = norm_pdf(xx, self.flen_mean, self.flen_std)
            pl.fill(xx, yy, 'k')#, linewidth=2.0)
            pl.xlabel("fragment length")
            pl.ylabel("$p(L)$")
            pl.xlim(0, 400)

        #position bias
        if mode == "pos5" or mode == "pos3":
            pl.plot(np.arange(20), np.ones(20), '--k')
            for i in range(5):
                _label="bin%d: %.0f-%.0f bp" %(i+1, self.percentile[i,0], self.percentile[i,1])
                if mode == "pos5":
                    pl.plot(np.arange(20)+0.5, self.pos5_prob[i,:], linewidth=2.0, label=_label)
                else:
                    pl.plot(np.arange(20)+0.5, self.pos3_prob[i,:], linewidth=2.0, label=_label)
            pl.legend(loc="best")
            pl.xlabel("fractional transcription position")
            pl.ylabel("bias weight")
            pl.ylim(0,2)

        #sequence bias
        if mode == "seq5" or mode == "seq3":
            base = ["A", "T", "G", "C"]
            _color = ["g", "r", "orange", "b"]
            if mode == "seq5":
                pl.plot(np.arange(21)-8, np.ones(21), '--k')
                pl.plot(np.zeros(2), np.array([0, 2.0]), '--k', linewidth=2.0)
                percent = np.zeros((4,21))
                for i in range(4):
                    for j in range(21):
                        _seq_bias = self.seq5_prob[str(j)]
                        percent[i,j] = np.sum(_seq_bias[i*4**(self.chain_len[j]-1): 
                            (i+1)*4**(self.chain_len[j]-1)]) / 4**(self.chain_len[j]-1)
                    pl.plot(np.arange(21)-8, percent[i,:], ":o", c=_color[i], label=base[i])
                pl.xlabel("offset from 3' fragment end")
                pl.xlim(-8,12)
            else:
                pl.plot(np.arange(21)-12, np.ones(21), '--k')
                pl.plot(np.zeros(2), np.array([0, 2.0]), '--k', linewidth=2.0)
                percent = np.zeros((4,21))
                for i in range(4):
                    for k in range(21):
                        j = 20 - k
                        _seq_bias = self.seq3_prob[str(j)]
                        percent[i,j] = np.sum(_seq_bias[i*4**(self.chain_len[j]-1): 
                            (i+1)*4**(self.chain_len[j]-1)]) / 4**(self.chain_len[j]-1)
                    pl.plot(np.arange(21)-12, percent[i,:], ":o", c=_color[i], label=base[i])
                pl.xlabel("offset from 3' fragment end")
                pl.xlim(-12,8)
            pl.legend(loc="best")
            pl.xlabel("offset from %s' fragment end" %mode[3])
            pl.ylabel("bias weight")
            pl.ylim(0.5,2)

        #legend only
        if mode == "legend":
            base = ["A", "T", "G", "C"]
            _color = ["g", "r", "orange", "b"]
            pl.axis('off')
            ax1 = pl.twinx()
            for i in range(len(base)):
                ax1.plot([], [], "o", c=_color[i], label=base[i])
            ax1.legend(numpoints=1, loc=4)
            ax1.axis('off')
            ax2 = pl.twinx()
            for i in range(5):
                _label="bin%d: %.0f-%.0f bp" %(i+1, self.percentile[i,0], self.percentile[i,1])
                ax2.plot([], [], linewidth=2.0, label=_label)
            ax2.legend(loc=3)
            ax2.axis('off')


# This module is to process the reads, and the location index of reads 
# in each state, i.e.,  exons, juntion between exons.

import numpy as np

class ReadSet(object):
    """docstring for ReadSet"""
    def __init__(self, reads1, reads2=[]):
        self.length = len(reads1)
        if len(reads1) != len(reads2): self.size = 1
        if len(reads1) == len(reads2): self.size = 2

        self.uloc = np.ones((self.length,self.size)) #up-stream location(small)
        self.dloc = np.ones((self.length,self.size)) #down-stream location(big)
        self.qlen = np.ones((self.length,self.size)) #aligned reads length
        self.rlen = np.ones((self.length,self.size)) #orignal reads length
        self.prob = np.ones((self.length,self.size)) #reads report probability
        self.mapp = np.ones((self.length,self.size)) #reads mapping probability

        for s in range(self.size):
            if s == 0: reads = reads1
            if s == 1: reads = reads2
            for i in range(self.length):
                r = reads[i]
                self.uloc[i,s] = r.pos
                self.dloc[i,s] = r.aend - 1 # qlen=dloc-uloc-1
                self.qlen[i,s] = r.qlen
                self.rlen[i,s] = r.rlen
                self.mapp[i,s] = 1.0 - 10 ** (0 - r.mapq / 10.0)
                qual_bases = []
                qual_bases[:] = r.qqual
                #Note: still need improvement on the mismatched bases
                for _qbase in qual_bases:
                    self.prob[i,s] = self.prob[i,s]*(1-10**(-ord(_qbase)/10.0))
                # if r.rlen < (r.aend - r.pos): print(r.rlen, r.aend, r.pos)

    def get_loc_idx(self, exons, strand, overhang=1):
        """get the index on exon or intron"""
        self.strand = strand
        self.exons  = np.sort(np.array(exons).reshape(-1,2), axis=0)
        self.seglen = np.array([self.exons[0,1]-self.exons[0,0] + 1,
                                self.exons[1,0]-self.exons[0,1] - 1,
                                self.exons[1,1]-self.exons[1,0] + 1])
        
        loc_idx = np.zeros((self.length, 8), bool)
        len_use = np.zeros((self.length, 8))

        flen  = self.dloc.max(axis=1) - self.uloc.min(axis=1) + 1
        rlen1 = self.dloc.min(axis=1) - self.uloc.min(axis=1) + 1
        rlen2 = self.dloc.max(axis=1) - self.uloc.max(axis=1) + 1

        # exon1
        loc_idx[:,0] = self.dloc.max(axis=1) <= self.exons[0,1]
        len_use[:,0] = self.seglen[0] - flen + 1

        # exon1-intron1 boundary, SSite5
        loc_idx[:,1] = ((self.uloc.min(axis=1) <= self.exons[0,1]-overhang+1) *
                        (self.dloc.max(axis=1) <= self.exons[1,0]-1) *
                        (self.dloc.max(axis=1) >= self.exons[0,1]+overhang))
        len_use[:,1] = np.minimum(flen, min(self.seglen[0], self.seglen[1]))

        # intron1
        loc_idx[:,2] = ((self.uloc.min(axis=1) >= self.exons[0,1]+1) *
                        (self.dloc.max(axis=1) <= self.exons[1,0]-1))
        len_use[:,2] = self.seglen[1] - flen + 1

        # intron1-exon2 boundary, SSite3
        loc_idx[:,3] = ((self.uloc.min(axis=1) >= self.exons[0,1]+1) *
                        (self.uloc.min(axis=1) <= self.exons[1,0]-overhang) *
                        (self.dloc.max(axis=1) >= self.exons[1,0]+overhang-1))
        len_use[:,3] = np.minimum(flen, min(self.seglen[1], self.seglen[2]))

        # exon2
        loc_idx[:,4] = self.uloc.min(axis=1) >= self.exons[1,0]
        len_use[:,4] = self.seglen[2] - flen + 1

        # exon1-exon2 junction, junction
        loc_idx[:,5] = ((self.uloc <= self.exons[0,1]-overhang+1) *
                        (self.dloc >= self.exons[1,0]+overhang-1) *
                        (self.dloc-self.uloc-self.qlen > self.seglen[1]-5)
                        ).sum(axis=1)>0
        len_use[:,5] = np.minimum(np.maximum(rlen1, rlen2), 
                                  min(self.seglen[0], self.seglen[2]))

        # exon1-intron-exon2
        loc_idx[:,6] = ((self.uloc.min(axis=1) <= self.exons[0,1]) *
                        (self.dloc.max(axis=1) >= self.exons[1,0]) *
                        ((self.uloc.max(axis=1) <= self.exons[1,0]-overhang) +
                         (self.dloc.min(axis=1) >= self.exons[0,1]+overhang))*
                        (True - loc_idx[:,5]))
        len_use[:,6] = min(self.seglen[0], self.seglen[2])
        
        # exon1-exon2 unsure, exon_both
        loc_idx[:,7] = ((self.uloc.max(axis=1) >= self.exons[1,0]) *
                        (self.dloc.min(axis=1) <= self.exons[0,1]))
        len_use[:,7] = np.minimum(self.seglen[0]-rlen1,self.seglen[2]-rlen2)+1

        idx = np.where(len_use <= 0)
        len_use[idx] = 1.0

        # revers strand
        if ["-","-1","0",0,-1].count(self.strand) > 0: 
            self.seglen = self.seglen[::-1]
            for i in range(loc_idx.shape[0]):
                loc_idx[i,:5] = loc_idx[i,:5][::-1]
                len_use[i,:5] = len_use[i,:5][::-1]

        self.loc_idx  = loc_idx
        self.len_use  = len_use
        self.RPK_use  = loc_idx / len_use
        self.loc_name = ["exon1", "SSite5", "intron", "SSite3", "exon2",
                         "jucntion", "ex1_intn_ex2", "ex1_ex2_unsure"]

        # Rmat and flen are both for pre-mRNA (p) and mature mRNA (m)
        self.Rmat = np.zeros((self.length, 2), dtype="bool")
        self.Rmat[:,0] = (self.loc_idx[:,:5].sum(axis=1) + 
                          self.loc_idx[:,6:].sum(axis=1)) #p
        self.Rmat[:,1] = (self.loc_idx[:,0] + self.loc_idx[:,4] +
                          self.loc_idx[:,5] + self.loc_idx[:,7]) #m

        idx = self.loc_idx[:,5] + self.loc_idx[:,7] #junctions
        self.flen = np.zeros((self.length, 2))
        self.flen[:, 0]   = self.dloc.max(axis=1) - self.uloc.min(axis=1)
        self.flen[:, 1]   = self.dloc.max(axis=1) - self.uloc.min(axis=1)
        self.flen[idx, 1] = self.flen[idx, 1] - self.seglen[1]

        


# This file is to esitmate the sequecne and positon bias parameters.

import sys
import numpy as np
from optparse import OptionParser
from .utils.tran_utils import TranUnits
from .utils.gtf_utils import load_annotation
from .utils.bias_utils import BiasFile, FastaFile
from .utils.sam_utils import load_samfile, fetch_reads


def get_bias(BF, transcript, ref_seq, reads, threshold=0.01):
    """to get the bias parameters from a transcript `t`."""
    rcount = len(reads["reads1"])+len(reads["reads1u"])+len(reads["reads2u"])
    if rcount < threshold * transcript.tranL:
        print("Coverage is only RPK=%.1f on %s, skipped." 
            %(rcount * 1000.0 / transcript.tranL, transcript.tranID))
        return BF, np.array([], "float")
    t = TranUnits(transcript)
    t.set_sequence(ref_seq)
    seqs = t.seq
    tLen = t.ulen
    flen = np.array([], "float")
    idx5 = np.array([], "float")
    idx3 = np.array([], "float")

    if len(reads["reads1"]) > 0:
        t.set_reads(reads["reads1"], reads["reads2"], "unif")
        idx5 = np.append(idx5, t.idx5)
        idx3 = np.append(idx3, t.idx3)
        flen = np.append(flen, t.flen[t.Rmat])

    if len(reads["reads1u"]) > 0:
        t.set_reads(reads["reads1u"], [], "unif")
        idx5 = np.append(idx5, t.idx5)

    if len(reads["reads2u"]) >0:
        t.set_reads([], reads["reads2u"], "unif")
        idx3 = np.append(idx3, t.idx3)
    
    _i5 = idx5 == idx5
    _i3 = idx3 == idx3
    idx5 = idx5[_i5]
    idx3 = idx3[_i3]

    for i in range(tLen):
        ipos = i + 20
        _seq5 = t.seq[ipos-8  : ipos+13]
        _seq3 = t.seq[ipos-12 : ipos+9 ][::-1]
        _pos5 = i
        _pos3 = i #tLen - i -1
        if t.strand != "+" and t.strand != "1":
            _seq5, _seq3 = _seq3, _seq5
            # _pos5, _pos3 = _pos3, _pos5
            _pos5, _pos3 = tLen-1-_pos3, tLen-1-_pos5
        if len(idx5) > 0 and np.mean(idx5==i) > 0:
            BF.set_both_bias(_seq5, _pos5, tLen, np.mean(idx5==i), 5, "bias")
        if len(idx3) > 0 and np.mean(idx3==i) > 0: 
            BF.set_both_bias(_seq3, _pos3, tLen, np.mean(idx3==i), 3, "bias")
        BF.set_both_bias(_seq5, _pos5, tLen, 1.0/tLen, 5, "unif")
        BF.set_both_bias(_seq3, _pos3, tLen, 1.0/tLen, 3, "unif")

    return BF, flen

def main():
    # import warnings
    # warnings.filterwarnings('error')

    #part 0. parse command line options
    parser = OptionParser()
    parser.add_option("--anno_file", "-a", dest="anno_file", default=None,
        help="The annotation file in gtf format")
    parser.add_option("--anno_source", dest="anno_source", default="Ensembl",
        help="The annotation source of the gtf file")
    parser.add_option("--sam_file", "-s", dest="sam_file", default=None,
        help="The indexed alignement file in bam/sam format")
    parser.add_option("--refseq_file", "-r", dest="refseq_file", default=None,
        help="The fasta file of genome reference sequences")
    parser.add_option("--out_file", "-o", dest="out_file",
        default="parameters.hdf5", help="The output file in hdf5 format")

    (options, args) = parser.parse_args()
    if options.anno_file == None:
        print("Error: need --anno_file for annotation.")
        sys.exit(1)
    else:
        anno = load_annotation(options.anno_file, options.anno_source)
    if options.sam_file == None:
        print("Error: need --sam_file for reads indexed and aliged reads.")
        sys.exit(1)
    else:
        samFile = load_samfile(options.sam_file)

    biasFile = BiasFile()
    out_file = options.out_file
    fastaFile = FastaFile(options.refseq_file)

    #part 1.1. all transcript length
    genes = anno["genes"]
    tran_len_all = []
    for g in genes:
        for t in g.trans:
            tran_len_all.append(t.tranL)
    biasFile.set_percentile(np.array(tran_len_all))
    print(biasFile.percentile)
    print("%.0f out of %.0f genes have one isoform for bias parameter estimate."
          %(np.sum(anno["tran_num"]==1), len(anno["tran_num"])))

    import time
    start_time = time.time()

    cnt = 0
    flen_all = np.array([],"float")
    for g in genes:
        if g.tranNum > 1: continue
        cnt += 1
        print(cnt, g.geneID, g.chrom, g.strand, g.start, g.stop)
        
        # 2.1 get the reads and seg_len
        reads = fetch_reads(samFile, g.chrom, g.start, g.stop, 
                            rm_duplicate=True, inner_only=True, mapq_min=10,
                            mismatch_max=10, rlen_min=1, is_mated=True)
        if len(reads["reads1"])+len(reads["reads1u"])+len(reads["reads2u"])==0:
            print("No reads for %s" %g.geneID)
            continue
        else:
            print(len(reads["reads1"]), len(reads["reads1u"]), len(reads["reads2u"]))

        biasFile, _flen = get_bias(biasFile, g.trans[0], fastaFile, reads)
        flen_all = np.append(flen_all, _flen)

    biasFile.flen_mean = np.mean(flen_all)
    biasFile.flen_std = np.std(flen_all)
    biasFile.save_file(out_file)

    print("--- %.3f seconds ---" % (time.time() - start_time))
    

if __name__ == "__main__":
    main()



import numpy as np

def logistic_normal_pdf(D, mu, cov, log=False):
    """
    Calculate the probability density of logistic-normal distribution, see
    Atchison J, Shen S M. Logistic-normal distributions: Some properties and 
    uses[J]. Biometrika, 1980, 67(2): 261-272.
    http://www.jstor.org/stable/2335470

    Parameters
    ----------
    D   : 1-D array_like, (K, )
        The variable for calculating the probability density. Range: [0,1].
    mu  : 1-D array_like, (K-1, )
        Mean of the logistic multinormal distribution. Range: [-inf,inf].
    cov : 1-D array_like, (K-1, K-1)
        Covariance of for the logistic multinormal distribution.
    log : bool
        If true, the return value is at log scale.

    Returns
    -------
    RV  : numpy float
        The probability density of D.
    """
    if len(D.shape) > 1:
        print("Only one array is calculated each time.")
        return None
    D_logit = np.log(D[:-1]/D[-1]) - mu
    cov_inv = np.linalg.inv(cov)
    cov_det = np.linalg.det(cov)
    if cov_det < 0:
        print("The det of covariance is negative, please check!")
        return None
    RV = (-0.5*np.log(2*np.pi*cov_det) - np.sum(np.log(D)) - 
           0.5*np.dot(np.dot(D_logit, cov_inv), D_logit))
    if log == False: RV = np.exp(RV)
    return RV


def Dirichlet_pdf(D, alpha, log=False):
    """
    Calculate the probability density of Dirichlet distribution.

    Parameters
    ----------
    D : 1-D array_like, (K, )
        The variable for calculating the probability density. Range: [0,1].
    alpha : 1-D array_like, (K, )
        The parameter array for the Dirichlet distribution.
    log : bool
        If true, the return value is at log scale.

    Returns
    -------
    RV : numpy float
        The probability density of D.
    """
    if len(D.shape) > 1:
        print("Only one array is calculated each time.")
        return None
    part2 = 0
    for i in range(len(alpha)):
        part2 += np.math.lgamma(alpha[i])
    RV = np.math.lgamma(np.sum(alpha)) - part2 + np.sum((alpha-1)*np.log(D))
    if log == False: RV = np.exp(RV)
    return RV

def Geweke_Z(X, first=0.1, last=0.5):
    """
    Geweke diagnostics for MCMC chain convergence.
    See Geweke J. Evaluating the accuracy of sampling-based approaches to the 
    calculation of posterior moments[M]. Minneapolis, MN, USA: Federal Reserve 
    Bank of Minneapolis, Research Department, 1991.
    and https://pymc-devs.github.io/pymc/modelchecking.html#formal-methods

    Parameters
    ----------
    X : 1-D array_like, (N, )
        The uni-variate MCMC sampled chain for convergence diagnostic.
    first : float
        The proportion of first part in Geweke diagnostics.
    last : float
        The proportion of last part in Geweke diagnostics.

    Returns
    -------
    Z : float
        The Z score of Geweke diagnostics.
    """
    N = X.shape[0]
    A = X[:first*N]
    B = X[last*N:]
    Z = abs(A.mean() - B.mean()) / np.sqrt(np.var(A) + np.var(B))
    return Z

def Psi_MCMC_MH(R_mat, len_isos, prob_isos, cov_fixed, alpha_Dir, M=100000, start=3000, gap=1000):
    """
    Calculate the proportion of each isoforms with all reads via MCMC samplling
    Parameters
    ----------
    R_mat : 2-D array_like, (N, K)
        N reads identities of belonging to K isoforms.
    len_isos : 1-D array_like, (K,)
        The effective length of each isoform.
    cov_fixed : 2-D array_like, (K-1, K-1)
        The fixed covariance of Logistic-Normal distribution.
    alpha_Dir : 1-D array_like, (K,)
        The parameter of Dirichlet distribution.
    M : int
        The iteration times of MH sampler, default=100.

    Returns
    -------
    RV : 2-D array_like, (M, K)
        M sampled Psi for abundance of each isoforms
    """
    # 1. check input data
    idx = (len_isos != len_isos)
    len_isos[idx] = 0.0
    prob_isos[:,idx] = 0.0
    R_mat[:,idx] = False

    idx = np.where(R_mat != R_mat)
    R_mat[idx] = False

    idx = np.where(prob_isos != prob_isos)
    prob_isos[idx] = 0.0

    idx = (R_mat.sum(axis=1) > 0) * (prob_isos.sum(axis=1) > 0)
    R_mat = R_mat[idx,:]
    prob_isos = prob_isos[idx,:]

    N, K = R_mat.shape

    # MCMC random initializations
    mu_now  = np.random.rand(K-1) - 0.5
    mu_now  = np.random.multivariate_normal(mu_now, cov_fixed)
    psi_now = np.zeros(K)
    psi_now[:-1] = np.exp(mu_now) / (1 + np.sum(np.exp(mu_now)))
    psi_now[-1]  = 1-np.sum(psi_now[:-1])
    fsi_now = len_isos*psi_now / np.sum(len_isos*psi_now)
    P_now   = (np.sum(np.log(np.dot(R_mat*prob_isos, fsi_now))) + 
               Dirichlet_pdf(psi_now, alpha_Dir, log=True))
    
    # MCMC running
    psi_try = np.zeros(K)
    Psi_all = np.zeros((M, K))
    for i in range(M):
        # step 1: propose a value
        mu_try  = np.random.multivariate_normal(mu_now, cov_fixed)
        psi_try[:-1] = np.exp(mu_try) / (1.0 + np.sum(np.exp(mu_try)))
        psi_try[-1]  = 1 - np.sum(psi_try[:-1])
        fsi_try = len_isos*psi_try / np.sum(len_isos*psi_try)
        P_try   = (np.sum(np.log(np.dot(R_mat*prob_isos, fsi_try))) + 
                   Dirichlet_pdf(psi_try, alpha_Dir, log=True))
        
        Q_now = logistic_normal_pdf(psi_now, mu_try, cov_fixed, log=True)
        Q_try = logistic_normal_pdf(psi_try, mu_now, cov_fixed, log=True)

        # step 2: calculate the MH ratio, accept or reject the proposal
        alpha = min(np.exp(P_try+Q_now-P_now-Q_try), 1)
        if alpha is None: print("alpha is none!")
        elif np.random.rand(1) < alpha:
            mu_now  = mu_try  + 0
            psi_now = psi_try + 0
            fsi_now = fsi_try + 0
            P_now   = P_try   + 0
        Psi_all[i,:] = psi_now

        #step 3. convergence diagnostics
        if i >= start and i % gap == 0:
            conv = 1
            for k in range(K):
                if Geweke_Z(Psi_all[:i, k]) > 2: 
                    conv = 0
                    break
            if conv == 1:
                Psi_all = Psi_all[:i,:]
                break

    return Psi_all


def Psi_junction(R_mat, len_juncts):
    """
    Calculate the proportion of each isoforms with junction reads only
    Parameters
    ----------
    R_mat : 2D matrix, (N, K)
        Parameters; N and K are the number of reads and isoforms, respectively.
    len_juncts : 1D array, (K,)
        Parameters; the used junction length of each isoforms
    Returns
    -------
    RV : 1D array, (K,)
    The the proportion of each isoforms
    """
    N, K = R_mat.shape
    if R_mat.shape[1] != len_juncts.shape[0]:
        print("The number of isoforms in the reads and junction lengths are different!")
        return [None] * K
    if np.sum(R_mat) == 0:
        print("No junction reads!")
        return [None] * K
    Dens = np.sum(R_mat, axis=0) / len_juncts.astype("float")
    RV = Dens / np.sum(Dens)
    return RV

def Psi_analytic(R_mat, len_isos):
    """
    Calculate the proportion of each isoforms with all reads
    Parameters
    ----------
    R_mat : 2D matrix, (N, K)
        Parameters; N and K are the number of reads and isoforms, respectively.
    len_isos : 1D array, (K,)
        Parameters; the used length of each isoforms
    Returns
    -------
    RV : 1D array, (K,)
    The the proportion of each isoforms
    """
    # transfer the parameters
    p_p, p_m = 1.0/len_isos

    # counting the reads of each type 
    Nc = np.sum(R_mat[:,0] * R_mat[:,1])
    Np = np.sum(R_mat[:,0] * (R_mat[:,1]==False))
    Nm = np.sum(R_mat[:,1] * (R_mat[:,0]==False))

    # The coefficients of a quadratic equation
    A = (-Np-Nm-Nc) * (p_p-p_m)
    B = Np*(p_p - 2*p_m) - Nm*p_m + Nc*(p_p - p_m)
    C = Np * p_m

    # # The analytical solution the quadratic equation on psi_freq
    Psi_freq = np.zeros(2)
    Psi_freq[0] = (-B - np.sqrt(B**2-4*A*C)) / (2.0*A)
    Psi_freq[1] = 1 - Psi_freq[0]
    RV = Psi_freq*len_isos[::-1] / np.sum(Psi_freq*len_isos[::-1])
    return RV




# This is a direct running file to calculate the specific counts of RNA
# splicing, and currently it is mainly for intronic splicing, rather than
# alternative splicing. There are 8 types for reads for an exon1-intron-
# exon2 structure: (0) exon1, (1) exon1-intron boundary, (2) intron, (3) 
# intron-exon2 boundary, (4) exon2, (5) exon1-exon2 junction, (6) exon1-
# intron-exon2, (7) exon1-exon2 unsure. In addition, it also provides the
# normalized counts RPK (reads per kilo-base).

import sys
import h5py
import pysam
import numpy as np
from optparse import OptionParser
from .utils.reads_utils import ReadSet
from .utils.gtf_utils import load_annotation
from .utils.sam_utils import load_samfile, fetch_reads

def main():
    print("Welcome to dice-count!")

    #part 0. parse command line options
    parser = OptionParser()
    parser.add_option("--anno_file", "-a", dest="anno_file", default=None,
        help="The annotation file in gtf format")
    parser.add_option("--anno_source", dest="anno_source", default="Ensembl",
        help="The annotation source of the gtf file")
    parser.add_option("--sam_file", "-s", dest="sam_file", default=None,
        help="The indexed alignement file in bam/sam format")
    parser.add_option("--total_reads", dest="total_reads", default=None,
        help="The total reads for calculate RPKM.")
    parser.add_option("--gene_file", "-g", dest="gene_file",
        help="The list of genes in use.")
    parser.add_option("--out_file", "-o", dest="out_file",
        default="dice_count.txt", help="The counts in plain text file")
    parser.add_option("--rm_duplicate", dest="rm_duplicate",
        default="True", help="remove duplicate reads or not.")
    parser.add_option("--inner_only", dest="inner_only",
        default="True", help="only unclude the reads inside or not.")
    parser.add_option("--mapq_min", dest="mapq_min",
        default="10", help="the minimum mapq for reads.")
    parser.add_option("--mismatch_max", dest="mismatch_max",
        default="5", help="the maximum mismatch for reads.")
    parser.add_option("--rlen_min", dest="rlen_min",
        default="1", help="the mimimum length of reads.")
    parser.add_option("--is_mated", dest="is_mated",
        default="True", help="process reads as paired-end or not.")
    parser.add_option("--total_only", dest="total_only", default="True",
        help="provide total reads count only; if False the specific reads" +
        " for the exon-intron-exon structure.")
    parser.add_option("--biotype_rm", dest="biotype_rm", default=None,
        help="The exclusive biotype.")
    parser.add_option("--biotype_only", dest="biotype_only", default=None,
        help="The only used biotype.")

    (options, args) = parser.parse_args()
    anno_source = options.anno_source
    if options.anno_file == None:
        print("Error: need --anno_file for annotation.")
        sys.exit(1)
    else:
        anno = load_annotation(options.anno_file, anno_source)
    if options.sam_file == None:
        print("Error: need --sam_file for reads indexed and aliged reads.")
        sys.exit(1)
    else:
        samFile = load_samfile(options.sam_file)

    if options.gene_file == None:
        gene_list = anno["gene_id"]
    else:
        gene_list = np.loadtxt(options.gene_file, dtype="str")
    
    out_file     = options.out_file
    is_mated     = bool(options.is_mated == "True")
    mapq_min     = int(options.mapq_min)
    rlen_min     = int(options.rlen_min)
    inner_only   = bool(options.inner_only == "True")
    total_only   = bool(options.total_only == "True")
    biotype_rm   = options.biotype_rm
    total_reads  = options.total_reads
    biotype_only = options.biotype_only
    rm_duplicate = bool(options.rm_duplicate == "True")
    mismatch_max = int(options.mismatch_max)

    cnt_name = ["ex1", "ex1_int", "int", "int_ex2", "ex2", "ex1_ex2_junc",
                "ex1_int_ex2", "ex1_ex2_vague"]

    if total_reads is None: RPKM_symbol = "_RPK"
    else: RPKM_symbol = "_RPKM"

    fid = open(out_file, "w")
    if total_only != True:
        head_line = "gene_id\tgene_name\tbiotype"
        for i in range(len(cnt_name)):
            head_line += "\t" + cnt_name[i] + "_NUM"
        for i in range(len(cnt_name)):
            head_line += "\t" + cnt_name[i] + RPKM_symbol
    else:
        head_line = "gene_id\tgene_name\tbiotype\tcount\t" + RPKM_symbol[1:]
    fid.writelines(head_line + "\n")
    
    g_cnt = 0
    for g in range(gene_list.shape[0]):
        i = np.where(np.array(anno["gene_id"]) == gene_list[g])[0][0]

        if (biotype_rm is not None and 
            biotype_rm.split("---").count(anno["biotype"][i]) == 1): continue
        if (biotype_only is not None and 
            biotype_only.split("---").count(anno["biotype"][i]) == 0): continue

        _gene   = anno["gene_id"][i]
        _strand = anno["strand"][i]
        _chrom  = str(anno["chrom"][i])
        _start  = anno["gene_start"][i]
        _stop   = anno["gene_stop"][i]

        if total_only == True:
            reads = fetch_reads(samFile, _chrom, _start, _stop, rm_duplicate,
                            inner_only, mapq_min, mismatch_max,
                            rlen_min, is_mated)

            count = len(reads["reads1u"]) + len(reads["reads2u"])
            count += len(reads["reads2"])
            if total_reads is None:
                RPK   = count * 1000.0 / abs(_stop - _start + 1)
            else:
                RPK   = count * 10**9 / (abs(_stop - _start + 1) * float(total_reads))
            count = [count]
            RPK   = [RPK]
            g_cnt += 1
        elif (anno["genes"][i].tranNum > 0 and 
              anno["genes"][i].trans[0].exonNum == 2):
            reads = fetch_reads(samFile, _chrom, _start, _stop, rm_duplicate,
                            inner_only, mapq_min, mismatch_max,
                            rlen_min, is_mated)

            rdSet = ReadSet(reads["reads1u"])
            rdSet.get_loc_idx(anno["genes"][i].trans[0].exons, _strand)
            count = rdSet.loc_idx.sum(axis=0)
            RPK   = rdSet.RPK_use.sum(axis=0)

            rdSet = ReadSet(reads["reads2u"])
            rdSet.get_loc_idx(anno["genes"][i].trans[0].exons, _strand)
            count += rdSet.loc_idx.sum(axis=0)
            RPK   += rdSet.RPK_use.sum(axis=0)

            rdSet = ReadSet(reads["reads1"], reads["reads2"])
            rdSet.get_loc_idx(anno["genes"][i].trans[0].exons, _strand)
            count += rdSet.loc_idx.sum(axis=0)
            RPK   += rdSet.RPK_use.sum(axis=0)

            if total_reads is not None:
                RPK = RPK * 10**6 / float(total_reads)
            g_cnt += 1
        else: continue

        a_line = anno["gene_id"][i] + "\t" + anno["gene_name"][i] + "\t"
        a_line += anno["biotype"][i] + "\t"
        a_line += "\t".join(["%d" %num for num in list(count)]) + "\t" 
        a_line += "\t".join(["%.2f" %num for num in list(RPK)]) + "\n"
        fid.writelines(a_line)

        if g_cnt % 100 == 0 and g_cnt != 0:
            print("%d genes have been processed." %g_cnt)
    fid.close()
    print("%d genes have been processed. Done!" %g_cnt)

if __name__ == "__main__":
    main()



# This is a main file to run the diceseq software, which will return the 
# isoform proportions ratio for each gene at all time points.

import sys
import gzip
import numpy as np
from optparse import OptionParser
from .models.model_GP import Psi_GP_MH, Psi_EM, EM_filter, EM_bootstrap
from .models.model_static import Psi_MCMC_MH
from .utils.gtf_utils import load_annotation
from .utils.bias_utils import BiasFile, FastaFile
from .utils.sam_utils import load_samfile, fetch_reads
from .utils.tran_utils import TranUnits, TranSplice

def get_CI(data, percent=0.95):
    if len(data.shape) == 0:
        data = data.reshape(-1,1)
    RV = np.zeros((data.shape[1],2))
    CI_idx = int(data.shape[0] * (1-percent)/2)
    for k in range(data.shape[1]):
        temp = np.sort(data[:,k])
        RV[k,:] = [temp[-CI_idx], temp[CI_idx]]
    return RV

def harmonic_mean_log(data, log_in=True, log_out=True):
    if log_in is not True:
        data = np.log(data)
    RV = np.log(len(data)) + np.min(data, axis=0)
    RV = RV - np.log(np.sum(np.exp(np.min(data, axis=0) - data), axis=0))
    if log_out is not True: RV = np.exp(RV)
    return RV

def Psi2Y(Psi):
    Psi = Psi / np.sum(Psi)
    Y = np.zeros(len(Psi))
    Y[:-1] = np.log(Psi[:-1] / Psi[-1])
    return Y

def main():
    import warnings
    warnings.filterwarnings('error')
    print("Welcome to diceseq GP model!")

    #part 0. parse command line options
    parser = OptionParser()
    parser.add_option("--anno_file", "-a", dest="anno_file", default=None,
        help="The annotation file in gtf format")
    parser.add_option("--anno_source", dest="anno_source", default="Ensembl",
        help="The annotation source of the gtf file")
    parser.add_option("--sam_list", "-s", dest="sam_list", default=None,
        help="the indexed alignement file in bam/sam format, use ',' for\
        replicates and '---' for time points, e.g.,\
        my_sam1_rep1.sorted.bam,my_sam1_rep2.sorted.bam---my_sam2.sorted.bam")
    parser.add_option("--time", "-t", dest="time", default=None,
        help="The time for the input samples, e.g., 0,1,2,3, the default\
        values will be the index of all time points, i.e., 0,1,...")
    parser.add_option("--ref_file", "-r", dest="ref_file", default=None,
        help="The genome reference file in fasta format. This is necessary\
        for bias correction, otherwise uniform mode will be used.")
    parser.add_option("--bias_file", "-b", dest="bias_file", default=None,
        help="The file for bias parameter")
    parser.add_option("--gene_file", "-g", dest="gene_file",
        help="The list genes in use. It is the gene id in the gtf annotation.")
    parser.add_option("--out_file", "-o", dest="out_file",
        default="diceseq_out", help="The prefix of the output file. There will\
        be two files: one in plain text format, the other in gzip format")
    # parser.add_option("--save_sample", dest="save_sample", default="True",
    #     help="whether save the sample file as well.")
    parser.add_option("--sample_num", dest="sample_num", default="500",
        help="The number of MCMC samples to save.")
    parser.add_option("--bias_mode", dest="bias_mode", default="unif", 
        help="The bias mode")
    parser.add_option("--add_premRNA", dest="add_premRNA", default="False",
        help="Whether adding pre-mRNA or not.")
    parser.add_option("--theta1_fix", dest="theta1_fix", default="3.0",
        help="The fixed hyperparameter theta1 for the GP model.")
    parser.add_option("--theta2_fix", dest="theta2_fix", default="None",
        help="The fixed hyperparameter theta2 for the GP model.")
    parser.add_option("--is_twice", dest="is_twice", default="True",
        help="Whether estimate the rates twice with a quick check first.")

    (options, args) = parser.parse_args()
    if options.anno_file == None:
        print("Error: need --anno_file for annotation.")
        sys.exit(1)
    else:
        anno = load_annotation(options.anno_file, options.anno_source)
    if options.sam_list == None:
        print("Error: need --sam_list for reads indexed and aliged reads.")
        sys.exit(1)
    else:
        sam_list = options.sam_list.split("---")
        samFiles = []
        for s in sam_list:
            _sam = []
            ss_list = s.split(",")
            for ss in ss_list:
                _sam.append(load_samfile(ss))
            samFiles.append(_sam)

    if options.gene_file == None:
        gene_list = anno["gene_id"]
    else:
        gene_list = np.loadtxt(options.gene_file, dtype="str")
    
    time      = options.time
    ref_file  = options.ref_file
    out_file  = options.out_file
    bias_file = options.bias_file
    bias_mode = options.bias_mode
    is_twice  = options.is_twice == "True"
    sample_num = int(options.sample_num)
    add_premRNA = options.add_premRNA == "True"
    # save_sample = options.save_sample == "True"
    theta1 = float(options.theta1_fix)
    try: theta2 = float(options.theta2_fix)
    except ValueError: theta2 = None

    if ref_file is None: 
        if bias_mode != "unif":
            bias_mode = "unif"
            print("No reference sequence, so we change to uniform mode.")
    else: fastaFile = FastaFile(ref_file)

    if bias_file is None: 
        if bias_mode != "unif":
            bias_mode = "unif"
            print("No bias parameter file, so we change to uniform mode.")
    else: biasFile = BiasFile(bias_file)

    # 1. run the model
    if time is None: X = np.arange(len(sam_list))
    else: X = np.array(time.split(","), "float")
    if theta2 is None: theta2 = ((max(X) - min(X) + 0.1) / 3.0)**2

    fid1 = open(out_file + ".dice", "w")
    headline = "gene_id\ttranscripts\ttransLen\tlogLik"
    for i in range(len(X)):
        _t = str(X[i])
        headline += "\tcount_T%s\tratio_T%s\t95CI_T%s" %(_t, _t, _t)
    fid1.writelines(headline + "\n")

    fid2 = gzip.open(out_file + ".sample.gz", "w")
    fid2.writelines("# MCMC samples for latent Y\n")
    fid2.writelines("# @gene|transcripts|theta2\n")
    fid2.writelines("# y_c1t1,y_c2t1;y_c1t2,y_c2t2\n")

    g_cnt = 0
    for g in range(gene_list.shape[0]):
        g_cnt += 1
        i = np.where(np.array(anno["gene_id"]) == gene_list[g])[0][0]

        if add_premRNA == True: 
            anno["genes"][i].add_premRNA()

        gene_info = anno["genes"][i].get_gene_info()
        _transLen = []
        for t in anno["genes"][i].trans:
            _transLen.append("%d" %t.tranL)
        transLen = ",".join(_transLen)

        R_all, len_iso_all, prob_iso_all, _count = [], [], [], []
        for j in range(len(samFiles)):
            t = TranSplice(anno["genes"][i])
            for _sam in samFiles[j]:
                t.set_reads(_sam)
            if bias_mode != "unif":
                t.set_sequence(fastaFile)
                t.set_bias(biasFile)

            t.get_ready(bias_mode)
            R_all.append(t.Rmat)
            if bias_mode == "unif": 
                len_iso_all.append(t.efflen_unif)
                prob_iso_all.append(t.proU)
            else: 
                len_iso_all.append(t.efflen_bias)
                prob_iso_all.append(t.proB)
            # _count.append(len(t.read1p) + len(t.read1u) + len(t.read2u))
            _count.append(np.sum(t.Rmat.sum(axis=1)>0))
        count = np.array(_count)
        
        print_info = ("%s: %d transcript, %d reads." %
            (anno["gene_id"][i], anno["genes"][i].tranNum, np.sum(count)))

        if R_all[0].shape[1] == 1:
            psi_mean = np.ones((1,len(X)))
            psi_95 = [np.ones((len(X), 2))]
            lik_marg = 0.0
            theta_mean = None
            sample_all = []

        elif R_all[0].shape[1] > 1:
            KP_FLAG, Psi_EM = EM_filter(R_all, len_iso_all, prob_iso_all, 2)

            for t in range(len(R_all)):
                R_all[t] = R_all[t][:,KP_FLAG]
                len_iso_all[t] = len_iso_all[t][KP_FLAG]
                prob_iso_all[t] = prob_iso_all[t][:,KP_FLAG]
            print_info += " KP_NUM: %d." %(sum(KP_FLAG))
            Ymean = np.zeros((R_all[0].shape[1], len(X)))
            for tt in range(len(X)):
                Ymean[:,tt] = Psi2Y(Psi_EM[tt, KP_FLAG] + 
                                    np.random.rand(np.sum(KP_FLAG)) * 0.0001)

            M = 20000
            gap = 100
            var = 0.1 * np.ones(np.sum(KP_FLAG)-1)
            initial = 1000 + gap * len(X)

            if is_twice == True:
                _var = np.zeros(R_all[0].shape[1]-1)
                for j in range(len(R_all)):
                    _Psi, _Y, _theta, _Pro, _Lik, _cnt, _m = Psi_GP_MH(R_all[j:j+1],
                        len_iso_all[j:j+1], prob_iso_all[j:j+1], X[j:j+1], 
                        Ymean[:,j:j+1], var, theta1, theta2, 100, 100, 50)
                    _var += np.var(_Y[int(_m/4):, :-1, 0], axis=0)
                var = (_var + 0.000000001) / len(R_all)
                #print(var)

            _Psi, _Y, _theta, _Pro, _Lik, _cnt, _m = Psi_GP_MH(R_all, len_iso_all, 
                prob_iso_all, X, Ymean, var, theta1, theta2, M, initial, gap)
            print_info += (" %d acceptances in %d iterations." %(_cnt, _m))

            psi_mean = np.zeros((len(KP_FLAG),len(X)))
            psi_95 = []
            _idx = -1
            for c in range(len(KP_FLAG)):
                if KP_FLAG[c]:
                    _idx += 1
                    psi_95.append(get_CI(_Psi[int(_m/4):,_idx,:], 0.95))
                else: 
                    psi_95.append(np.zeros((len(X), 2)))

            psi_mean[KP_FLAG,:] = _Psi[int(_m/4):,:,:].mean(axis=0)
            lik_marg = harmonic_mean_log(_Lik[int(_m/4):])
            theta_mean = _theta[int(_m/4):,:].mean(axis=0)
            sample_all = _Y[-sample_num:,:-1,:]

            _tran_name = []
            for tt in range(len(gene_info[-1].split(","))):
                if KP_FLAG[tt]: _tran_name.append(gene_info[-1].split(",")[tt])
            _theta2 = []
            for tt in range(len(theta_mean)):
                _theta2.append("%.2e" %_theta[int(_m/4):,tt].mean())
            _head = "@%s|%s|%s\n" %(gene_info[0], ",".join(_tran_name), ",".join(_theta2))
            _samples = ""
            for ss in range(sample_num):
                for tt in range(_Y.shape[2]):
                    _line_time = []
                    for cc in range(_Y.shape[1]):
                        _line_time.append("%.2e" %_Y[int(_m/2)+ss, cc, tt])
                    _samples += ",".join(_line_time)
                    if tt < _Y.shape[2] - 1: 
                        _samples += ";"
                    else:
                        _samples += "\n"
            fid2.writelines(_head)
            fid2.writelines(_samples)

        #print(print_info)

        aline = (gene_info[0] + "\t" + gene_info[-1] + "\t" + transLen + "\t%.1e" %lik_marg)
        for i in range(len(X)):
            _ratio,_ci95 = [], []
            for c in range(len(psi_mean[:,i])):
                _ratio.append("%.3f" %psi_mean[c,i])
                _ci95.append("%.3f:%.3f" %(psi_95[c][i,1], psi_95[c][i,0]))
            _ratio,_ci95 = ",".join(_ratio), ",".join(_ci95)
            aline += "\t%d\t%s\t%s" %(count[i], _ratio, _ci95)
        fid1.writelines(aline + "\n")

        if g_cnt % 10 == 0 and g_cnt != 0:
            print("%d genes have been processed." %g_cnt)
    print("%d genes have been processed. Done!" %g_cnt)
    fid1.close()
    fid2.close()


if __name__ == "__main__":
    main()
    


# This file is for generating reads for simulation

import sys
import gzip
import pysam
import numpy as np
from optparse import OptionParser
from diceseq.utils.tran_utils import TranUnits
from diceseq.utils.gtf_utils import load_annotation
from diceseq.utils.bias_utils import BiasFile, FastaFile

def GP_K(X, theta, x=None):
    N = len(X)
    if x is None:
        K = np.zeros((N, N))
        for i in range(N):
            for j in range(N):
                K[i,j] = theta[0] * np.exp(-0.5/theta[1]*(X[i]-X[j])**2)
    else:
        K = np.zeros(N)
        for i in range(N):
            K[i] = theta[0] * np.exp(-0.5/theta[1]*(X[i]-x)**2)
    return K

def YtoPsi(y):
    return np.exp(y) / (np.exp(y) + 1)

def YtoPsi(y, axis=0):
    """softmax function: transfer y to psi
    """
    size = np.array(y.shape)
    size[axis] = -1
    return np.exp(y) / (np.exp(y).sum(axis=axis).reshape(size))


def generate_reads(tran, N, rlen=100, flen=200, sigma=20, bias_mode="unif"):
    """generate the simulation reads. Only implimented the single-end
    reads for 5'end or 3'end. bias_mode could be 'unif', 'end5', 'end3'
    and 'both'. """
    reads1, reads2 = [], []
    if tran.strand == "1" or tran.strand == "+":
        idx5 = np.arange(0, tran.ulen-rlen+1)
        idx3 = np.arange(rlen-1, tran.ulen)
    else:
        idx5 = np.arange(rlen-1, tran.ulen)
        idx3 = np.arange(0, tran.ulen-rlen+1)

    cnt = -1
    probs = np.ones((len(idx5)+1)*len(idx5)/2)
    idx5s = np.ones((len(idx5)+1)*len(idx5)/2, "int")
    idx3s = np.ones((len(idx5)+1)*len(idx5)/2, "int")
    for i in range(len(idx5)):
        for j in range(i,len(idx3)):
            cnt += 1
            if tran.strand == "1" or tran.strand == "+":
                x = idx3[j] - idx5[i] + 1
                idx5s[cnt] = idx5[i]
                idx3s[cnt] = idx3[j]
            else:
                x = idx5[j] - idx3[i] + 1
                idx5s[cnt] = idx5[j]
                idx3s[cnt] = idx3[i]
            if x < 0: print("Errors! flen is negative!")
            probs[cnt] = normal_pdf(x, flen, sigma)
    if bias_mode == "end5" or bias_mode == "both":
        probs *= tran.bias5[idx5s]
    if bias_mode == "end3" or bias_mode == "both":
        probs *= tran.bias3[idx3s]
    probs = probs / np.sum(probs)
    count = np.random.multinomial(N, probs)

    for i in range(count.shape[0]):
        if count[i] == 0: continue
        if tran.strand == "1" or tran.strand == "+":
            fwd = tran.seq[idx5s[i]+20 : idx3s[i]+20+1]
        else:
            fwd = tran.seq[idx3s[i]+20 : idx5s[i]+20+1]
        rev = []
        rev[:] = fwd[::-1]
        for c in range(len(rev)):
            if   rev[c] == "A": rev[c] = "T"
            elif rev[c] == "T": rev[c] = "A"
            elif rev[c] == "G": rev[c] = "C"
            elif rev[c] == "C": rev[c] = "G"
        rev = "".join(rev)
        for j in range(count[i]):
            if tran.strand == "1" or tran.strand == "+":
                reads1.append(fwd[:rlen])
                reads2.append(rev[:rlen])
            else:
                reads1.append(rev[:rlen])
                reads2.append(fwd[:rlen])
    RV = {}
    RV["reads1"] = reads1
    RV["reads2"] = reads2
    return RV


def main():
    print "Welcome to dice-simulate. Reads are under generating."
    #part 0. parse command line options
    parser = OptionParser()
    parser.add_option("--anno_file", "-a", dest="anno_file", default=None,
        help="The annotation file in gtf format")
    parser.add_option("--anno_source", dest="anno_source", default="Ensembl",
        help="The annotation source of the gtf file: miso, ensemble")
    parser.add_option("--ref_file", "-r", dest="ref_file", default=None,
        help="The genome reference file in fasta format")
    parser.add_option("--bias_file", "-b", dest="bias_file", default=None,
        help="The parameter file for bias in hdf5 format")
    parser.add_option("--out_file", "-o", dest="out_file",
        default="simulate", help="The output file with simulated reads")
    parser.add_option("--add_premRNA", dest="add_premRNA", default="False",
        help="Whether adding pre-mRNA or not.")
    parser.add_option("--RPK", dest="RPK", default="1000", 
        help="The sequence depth, default=1000")
    parser.add_option("--ratio", dest="ratio", default="0.5", 
        help="The ratio of the first isoform, default=0.5.")
    parser.add_option("--times", dest="times", default=None, 
        help="The time points for generating reads with random ratios." + 
        "When it is not None, --ratio will be no useful, default=None.")
    parser.add_option("--noise", dest="noise", default="0",
        help="The noise in the reads number for each isoform, default=0.")
    parser.add_option("--rlen", dest="rlen", default="100", 
        help="The length of reads, default=100")
    parser.add_option("--fl_mean", dest="fl_mean", default="200", 
        help="The mean length of fragment, default=200")
    parser.add_option("--fl_sigma", dest="fl_sigma", default="20", 
        help="The stand variance of fragment length, default=20")
    parser.add_option("--bias_mode", dest="bias_mode", default="both", 
        help="The bias mode; unif, end5, end3, both")

    (options, args) = parser.parse_args()
    if options.anno_file == None:
        print "Error: need --anno_file for annotation."
        sys.exit(1)
    else:
        anno = load_annotation(options.anno_file, options.anno_source)

    if options.ref_file == None:
        print "Error: need --ref_file for reference sequence."
        sys.exit(1)
    else:
        fastaFile = FastaFile(options.ref_file)

    bias_mode = options.bias_mode
    if options.bias_file is None:
        biasFile = None
        if bias_mode != "unif":
            bias_mode = "unif"
            print "No bias parameter file, so we change to uniform mode."
    else: biasFile = BiasFile(options.bias_file)

    RPK = np.array(options.RPK.split(","), "int")
    if options.times is not None:
        T = np.arange(int(options.times))
    else: T = np.arange(0)
    Ymean = np.zeros(8)
    theta = np.array([3, 5.0])
    cov = GP_K(T, theta)
    #     ratio = YtoPsi(np.random.multivariate_normal(Ymean, cov))
    # else:
    #     ratio = np.array(options.ratio.split(","), "float")
    #     T = np.arange(len(ratio))
    # Np = np.zeros((len(RPK), len(ratio)), "int")
    # Nm = np.zeros((len(RPK), len(ratio)), "int")
    # cnt = np.zeros((len(RPK), len(ratio)), "int")

    rlen = int(options.rlen)
    noise = float(options.noise)
    fl_mean = float(options.fl_mean)
    fl_sigma = float(options.fl_sigma)
    out_file = options.out_file
    add_premRNA = options.add_premRNA == "True"

    fid0 = open(out_file + "_ratios.txt", "w")
    headline = "gene_id\ttrans_id"
    for t in range(len(T)):
        headline += "\tT%d" %T[t]
    fid0.writelines(headline + "\n")
    fid0.close()

    qua = []
    qua[:] = ">" * rlen
    qua = "".join(qua)
    for i in range(len(anno["gene_id"])):
        if i == 0: ftype = "w"
        else: ftype = "a+"
        _gene = anno["genes"][i]
        if add_premRNA == True: _gene.add_premRNA()

        unitp = TranUnits(anno["genes"][i].trans[1])
        unitm = TranUnits(anno["genes"][i].trans[0])
        unitp.set_sequence(fastaFile)
        unitm.set_sequence(fastaFile)
        unitp.set_bias(biasFile)
        unitm.set_bias(biasFile)

        if options.times is not None:
            ratio = YtoPsi(np.random.multivariate_normal(Ymean, cov))
        fid0 = open(out_file + "_ratios.txt", "a+")
        aline = _gene.geneID
        for t in range(len(T)):
            aline += "\t%.5f" %ratio[t]
        fid0.writelines(aline + "\n")
        fid0.close()

        for tr in range(len(_gene.trans)):

        for m in range(len(RPK)):
            for n in range(len(ratio)):
                _Np = max(0, (unitp.ulen) / 1000.0 * RPK[m] * (1-ratio[n]))
                _Nm = max(0, (unitm.ulen) / 1000.0 * RPK[m] * ratio[n])
                _aa = np.random.rand()
                _bb = np.random.rand()
                if _aa <= _Np - int(_Np): _Np = int(_Np) + 1.0
                else: _Np = int(_Np) + 0.0
                if _bb <= _Nm - int(_Nm): _Nm = int(_Nm) + 1.0
                else: _Nm = int(_Nm) + 0.0
                Np[m,n] = max(0, np.round(np.random.normal(_Np, max(0.01,noise*_Np))))
                Nm[m,n] = max(0, np.round(np.random.normal(_Nm, max(0.01,noise*_Nm))))
        readp = unitp.generate_reads(Np.max(),rlen,fl_mean,fl_sigma,bias_mode)
        readm = unitm.generate_reads(Nm.max(),rlen,fl_mean,fl_sigma,bias_mode)

        for m in range(len(RPK)):
            for n in range(len(ratio)):
                idxp = np.random.permutation(Np.max())[:Np[m,n]]
                idxm = np.random.permutation(Nm.max())[:Nm[m,n]]
                
                fid1 = gzip.open(out_file + "_RPK%d_T%d_1.fq.gz" %(RPK[m], n), ftype)
                fid2 = gzip.open(out_file + "_RPK%d_T%d_2.fq.gz" %(RPK[m], n), ftype)
                for r in idxp:
                    cnt[m,n] += 1
                    fid1.writelines("@simulated_reads_%d_1\n" %cnt[m,n])
                    fid2.writelines("@simulated_reads_%d_2\n" %cnt[m,n])
                    fid1.writelines(readp["reads1"][r] + "\n+\n" + qua + "\n")
                    fid2.writelines(readp["reads2"][r] + "\n+\n" + qua + "\n")
                for r in idxm:
                    cnt[m,n] += 1
                    fid1.writelines("@simulated_reads_%d_1\n" %cnt[m,n])
                    fid2.writelines("@simulated_reads_%d_2\n" %cnt[m,n])
                    fid1.writelines(readm["reads1"][r] + "\n+\n" + qua + "\n")
                    fid2.writelines(readm["reads2"][r] + "\n+\n" + qua + "\n")
                fid1.close()
                fid2.close()
        if i == 0 or (i+1) % 10 == 0 or (i+1) == len(anno["gene_id"]):
            print "%d out %d genes have been finished." % (i+1, len(anno["gene_id"]))

if __name__ == "__main__":
    main()




# This module is to process different transcript state, and set information
# for the transcript states, and finally get ready to run MCMC sampling.

import sys
import numpy as np
from .sam_utils import load_samfile, fetch_reads
from .bias_utils import FastaFile, BiasFile
from .gtf_utils import Gene, Transcript

def normal_pdf(x, mu, sigma):
    RV = 1 / (sigma*np.sqrt(2*np.pi)) * np.exp(-1.0/2*((x-mu)/sigma)**2)
    return RV

class TranUnits:
    """docstring for TranUnits"""
    def __init__(self, transcript):
        self.chrom  = transcript.chrom
        self.strand = transcript.strand
        self.units  = transcript.exons
        self.loci   = np.array([],"int")
        for i in range(self.units.shape[0]):
            _loci = np.arange(self.units[i,0], self.units[i,1]+1)
            self.loci = np.append(self.loci, _loci)
        self.ulen = len(self.loci)

    def set_sequence(self, fastaFile=None):
        """set the sequence of the transcript units, with 20 bases longer in 
        each end"""
        if fastaFile is not None:
            self.seq = fastaFile.get_seq(self.chrom, self.units[0,0] - 20, 
                                                     self.units[0,0] - 1)
            for i in range(self.units.shape[0]):
                self.seq += fastaFile.get_seq(self.chrom, self.units[i,0], 
                                                          self.units[i,1])
            self.seq += fastaFile.get_seq(self.chrom, self.units[i,1] + 1, 
                                                      self.units[i,1] + 20)
        else:
            print("This is a Null sequence file.")

    def set_bias(self, biasFile=None, mode="both"):
        """set the bias for all loci, with the bias modes of sequence / 
        position / both; make sure setting quence before using sequence
        or both modes"""
        if biasFile is None:
            # print("This is a Null bias file.")
            return
            
        self.bias_method = mode
        self.bias5 = np.ones(self.loci.shape[0])
        self.bias3 = np.ones(self.loci.shape[0])

        if ["seq", "sequence", "both"].count(mode) > 0:
            for i in range(len(self.bias5)):
                ipos = i + 20
                if self.strand == "+" or self.strand == "1":
                    _seq5 = self.seq[ipos-8  : ipos+13]
                    _seq3 = self.seq[ipos-12 : ipos+9 ][::-1]
                else:
                    _seq5 = self.seq[ipos-12 : ipos+9 ][::-1]
                    _seq3 = self.seq[ipos-8  : ipos+13]

                self.bias5[i] *= biasFile.get_seq_bias(_seq5, 5)
                self.bias3[i] *= biasFile.get_seq_bias(_seq3, 3)

        if ["pos", "position", "both"].count(mode) > 0:
            for i in range(len(self.bias5)):
                if self.strand == "+" or self.strand == "1":
                    _pos = i
                else:
                    _pos = len(self.loci) - 1 - i

                self.bias5[i] *= biasFile.get_pos_bias(_pos, self.ulen, 5)
                self.bias3[i] *= biasFile.get_pos_bias(_pos, self.ulen, 3)

    def get_index(self, loc):
        """get the location and the exon id of a loc on the transcript."""
        RV = [-1, -1]
        if   loc < self.units[0,0]:   RV[:] = [0, -2]
        elif loc > self.units[-1,-1]: RV[:] = [self.ulen-1, -3]
        else:
            cnt = 0
            for i in range(self.units.shape[0]):
                if loc >= self.units[i,0] and loc <= self.units[i,1]:
                    RV = [cnt + loc - self.units[i,0], i]
                    break
                else:
                    cnt += self.units[i,1] - self.units[i,0] + 1
        return RV

    def get_read_info(self, r1, r2, part_in=True):
        if r1 is None and r2 is None:
            return None
        idx5, idx3 = None, None
        mapq1, idx51, idx31 = 0.0, None, None
        mapq2, idx52, idx32 = 0.0, None, None
        if self.strand == "+" or self.strand == "1":
            if r1 is not None and r1.is_reverse:
                r1, r2 = r2, r1
            elif r2 is not None and r2.is_reverse == False:
                r1, r2 = r2, r1
        else:
            if r1 is not None and r1.is_reverse == False:
                r1, r2 = r2, r1
            elif r2 is not None and r2.is_reverse:
                r1, r2 = r2, r1

        if r1 is not None:
            mapq1 = 1.0 - 10 ** (0 - r1.mapq / 10.0)
            if self.strand == "+" or self.strand == "1":
                idx51 = self.get_index(r1.pos)
                idx31 = self.get_index(r1.aend - 1)
            else:
                idx31 = self.get_index(r1.pos)
                idx51 = self.get_index(r1.aend - 1)
            if idx51[1] == -1 or idx31[1] == -1: return None
            elif idx51[1] >= 0 and idx31[1] >= 0:
                if (abs(idx51[0] - idx31[0]) + 1 > r1.qlen + 3 or 
                    abs(idx51[0] - idx31[0]) + 1 < r1.qlen - 3): return None

        if r2 is not None:
            mapq2 = 1.0 - 10 ** (0 - r2.mapq / 10.0)
            if self.strand == "+" or self.strand == "1":
                idx52 = self.get_index(r2.pos)
                idx32 = self.get_index(r2.aend - 1)
            else:
                idx32 = self.get_index(r2.pos)
                idx52 = self.get_index(r2.aend - 1)
            if  idx52[1] == -1 or idx32[1] == -1: return None
            elif idx52[1] >= 0 and idx32[1] >= 0:
                if (abs(idx52[0] - idx32[0]) + 1 > r2.qlen + 3 or 
                    abs(idx52[0] - idx32[0]) + 1 < r2.qlen - 3): return None

        if r1 is None: 
            flen = abs(idx52[0] - idx32[0]) + 1
            if idx32[1] >= 0: idx3 = idx32[0]
        elif r2 is None:
            flen = abs(idx51[0] - idx31[0]) + 1
            if idx51[1] >= 0: idx5 = idx51[0]
        else:
            # flen = abs(idx32[0] - idx51[0]) + 1
            tmp1 = abs(idx32[0] - idx51[0]) + 1
            tmp2 = abs(idx31[0] - idx52[0]) + 1
            flen = max(tmp1, tmp2)

            # if self.strand == "+" and r1.is_reverse:
            #     print("test1")
            # if self.strand == "-" and r2.is_reverse:
            #     print("test2")

            # if abs(idx31[0] - idx52[0]) > abs(idx32[0] - idx51[0]): 
            #     if abs(idx51[0] - idx52[0]) < 5:
            #         print("strange!!!!!!", self.strand, r1.is_reverse, idx51[0], idx31[0], idx52[0], idx32[0])
            #     else:
            #         print("test", self.strand, r1.is_reverse, idx51[0], idx31[0], idx52[0], idx32[0])
            
            if idx51[1] >= 0: idx5 = idx51[0]
            if idx32[1] >= 0: idx3 = idx32[0]
            
        RV = {}
        RV["idx5"] = idx5
        RV["idx3"] = idx3
        RV["flen"] = flen
        RV["prob"] = max(mapq1, mapq2)
        return RV

    def set_reads(self, reads1=[], reads2=[], bias_mode="unif",  
        flen_mean=None, flen_std=None):
        """identify whether a read (pair) or is in this units, and return the 
        identity of the reads in this units, the units specific fragment  
        length bias scores. The bias score can be based on both (mode=both) 
        ends or single end5 (mode=end5) or single end3 (mode=end5), uniform 
        (mode=unif). Make sure the loc of read1 is smaller than read2."""
        if len(reads1) == 0 and len(reads2) == 0:
            print("Please input paired-end reads or singled end reads!")
            sys.exit(1)
        elif (len(reads1) * len(reads2)) != 0 and len(reads2) != len(reads1):
            print("Please input the same number of both mates of the reads!")
            sys.exit(1)

        self.rcnt = max(len(reads1), len(reads2))
        self.idx5 = np.ones(self.rcnt, "float")
        self.idx3 = np.ones(self.rcnt, "float")
        self.Rmat = np.ones(self.rcnt, "bool")
        self.flen = np.ones(self.rcnt, "float")
        self.proB = np.ones(self.rcnt, "float")
        self.proU = np.ones(self.rcnt, "float")
        self.bias_mode = bias_mode
        self.efflen_bias = 0
        self.efflen_unif = 0

        for i in range(self.rcnt):
            if len(reads1) == 0: r1 = None
            else: r1 = reads1[i]
            if len(reads2) == 0: r2 = None
            else: r2 = reads2[i]

            rinfo = self.get_read_info(r1, r2)
            if rinfo is None: 
                self.Rmat[i] = False
                self.flen[i] = None
                self.proB[i] = None
                self.proU[i] = None
                self.idx5[i] = None
                self.idx3[i] = None
            else:
                self.Rmat[i] = True
                self.flen[i] = rinfo["flen"]
                self.proB[i] = rinfo["prob"]
                self.proU[i] = rinfo["prob"]
                self.idx5[i] = rinfo["idx5"]
                self.idx3[i] = rinfo["idx3"]
                if self.bias_mode == "unif": continue
                elif self.bias_mode != "end3" and rinfo["idx5"] is not None:
                    self.proB[i] *= self.bias5[rinfo["idx5"]]
                elif self.bias_mode != "end5" and rinfo["idx3"] is not None:
                    self.proB[i] *= self.bias3[rinfo["idx3"]]

        # fragement distribution
        self.probs = np.zeros(self.ulen)
        if sum(self.Rmat) == 0: 
            self.probs[0] = 1.0
        elif np.unique(self.flen[self.Rmat]).shape[0] <= 10:
            for i in np.unique(self.flen[self.Rmat]):
                # self.probs[int(i)-1] = sum(self.flen[self.Rmat]==i) / (sum(self.Rmat) + 0.0)
                self.probs[int(i)-1] = np.mean(self.flen==i)
        else:
            if flen_std is None: flen_std = np.std(self.flen[self.Rmat])
            if flen_mean is None: flen_mean = np.mean(self.flen[self.Rmat])
            x = np.arange(1, self.ulen+1)
            self.probs[:] = normal_pdf(x, flen_mean, flen_std)
            self.probs = self.probs / sum(self.probs) #if sum(self.probs) != 0:
             
        # effective length
        self.biasLen = np.zeros(self.ulen)
        for i in range(1, self.ulen+1):
            self.efflen_unif += self.probs[i-1] * (self.ulen-i+1)
            if self.bias_mode == "unif" or self.probs[i-1] == 0: continue
            for j in range(self.ulen - i + 1):
                if self.strand == "+" or self.strand == "1":
                    pos5, pos3 = j, j+i-1
                else:
                    pos3, pos5 = j, j+i-1
                if   self.bias_mode == "end5": _bias = self.bias5[pos5]
                elif self.bias_mode == "end3": _bias = self.bias3[pos3]
                else : _bias = self.bias5[pos5] * self.bias3[pos3]
                self.biasLen[i-1] += _bias
            self.efflen_bias += self.probs[i-1] * self.biasLen[i-1]

        # reads probability
        for i in range(self.rcnt):
            if self.Rmat[i] == False: continue
            fL = int(self.flen[i])
            self.proU[i] *= self.probs[fL-1] / (self.ulen - fL + 1)
            if self.bias_mode != "unif":
                self.proB[i] *= (self.probs[fL-1] / self.biasLen[fL-1])

class TranSplice:
    def __init__(self, Gene):
        self.gene = Gene
        self.stop = Gene.stop
        self.chrom = Gene.chrom
        self.start = Gene.start
        self.unitSet = []
        for i in range(len(self.gene.trans)):
            self.unitSet.append(TranUnits(self.gene.trans[i]))
        self.read1p = []
        self.read2p = []
        self.read1u = []
        self.read2u = []

    def set_sequence(self, fastafile):
        """get the sequence from the genome sequence in FastaFile object"""
        for i in len(self.unitSet):
            self.unitSet[i].set_sequence(fastafile)

    def set_bias(self, biasfile):
        """get the bias parameters from the BiasFile object"""
        for i in len(self.unitSet):
            self.unitSet[i].set_bias(biasfile)
        self.bias_in = True

    def set_reads(self, samfile, rm_duplicate=True, inner_only=True,
                  mapq_min=0, mismatch_max=5, rlen_min=1, is_mated=True):
        """fetch the reads on this transcript from sam file"""
        reads = fetch_reads(samfile, self.chrom, self.start, self.stop,  
                            rm_duplicate, inner_only, mapq_min, 
                            mismatch_max, rlen_min,   is_mated)

        self.read1p += reads["reads1"]
        self.read2p += reads["reads2"]
        self.read1u += reads["reads1u"]
        self.read2u += reads["reads2u"]

    def get_ready(self, bias_mode="unif"):
        """get the location index of the transcript, need implimentation
        in future. Then, we could remove the Rmat, and flen from ReadSet
        object, and set the Rmat and flen by the info of states."""
        rcnt = [len(self.read1p), len(self.read1u), len(self.read2u)]
        unit_cnt = len(self.unitSet)
        self.Rmat = np.ones((sum(rcnt), unit_cnt), "bool")
        self.flen = np.ones((sum(rcnt), unit_cnt), "float")
        self.proB = np.ones((sum(rcnt), unit_cnt), "float")
        self.proU = np.ones((sum(rcnt), unit_cnt), "float")
        self.efflen_bias = np.zeros(unit_cnt, "float")
        self.efflen_unif = np.zeros(unit_cnt, "float")

        for i in range(unit_cnt):
            _units = self.unitSet[i]
            for j in range(3):
                _idx = np.arange(sum(rcnt[:j]), sum(rcnt[:j+1]))
                _reads1, _reads2 = [], []
                if   j == 0: _reads1, _reads2 = self.read1p, self.read2p
                elif j == 1: _reads1 = self.read1u
                elif j == 2: _reads2 = self.read2u
                if (len(_reads1) + len(_reads2)) == 0: continue

                _units.set_reads(_reads1, _reads2, bias_mode)
                self.Rmat[_idx, i] = _units.Rmat
                self.flen[_idx, i] = _units.flen
                self.proB[_idx, i] = _units.proB
                self.proU[_idx, i] = _units.proU

                self.efflen_unif[i] += sum(_units.Rmat) * _units.efflen_unif
                self.efflen_bias[i] += sum(_units.Rmat) * _units.efflen_bias

            if sum(self.Rmat[:, i]) > 0:
                self.efflen_unif[i] /= sum(self.Rmat[:, i])
                self.efflen_bias[i] /= sum(self.Rmat[:, i])
            else:
                self.efflen_unif[i] = _units.ulen
                self.efflen_bias[i] = _units.ulen

    def quick_start(self, samfile=None, fastafile=None, biasfile=None):
        """update the state_read after change the transcript information or 
        adding or removing states"""
        if samfile   is not None: self.set_reads(samfile)
        if fastafile is not None: self.set_sequence(fastafile)
        if biasfile  is not None: self.set_bias(biasfile)
        print(len(self.read1p), len(self.read1u), len(self.read2u))

        self.get_ready()
        print("ready")
        # try:
        #     self.get_ready()
        # except NameError:
        #     print("Cannot get ready now; please input necessary info first!")




import numpy as np

def erf(x):
    """error function approximation, with maximum error 1.5e-7.
    """
    # save the sign of x
    if x >= 0: sign = 1
    else: sign = -1
    x = abs(x)
    # constants
    a1 =  0.254829592
    a2 = -0.284496736
    a3 =  1.421413741
    a4 = -1.453152027
    a5 =  1.061405429
    p  =  0.3275911
    # A&S formula 7.1.26
    t = 1.0/(1.0 + p*x)
    y = 1.0 - (((((a5*t + a4)*t) + a3)*t + a2)*t + a1)*t*np.exp(-x*x)
    return sign*y # erf(-x) = -erf(x)


def gamma_pdf(x, k, theta, log=True):
    """
    Calculate the probability density of Gamma distribution.

    Parameters
    ----------
    x : float
        The variable for calculating the probability density.
    k : float
        The shape of the Gamma distribution.
    theta : float
        The scale of the Gamma distribution.
    log : bool
        If true, the return value is at log scale.

    Returns
    -------
    pdf : numpy float
        The probability density of x.
    """
    if k == 0: 
        print("The shape of Gamma distribution cannot be 0, please check!")
        return None
    pdf = -np.math.lgamma(k) - k*np.log(theta)
    pdf += (k-1)*np.log(x) - x/theta
    if log == False: pdf = np.exp(pdf)
    return pdf

def normal_pdf(x, mu, cov, log=True):
    """
    Calculate the probability density of Gaussian (Normal) distribution.

    Parameters
    ----------
    x : float or 1-D array_like, (K, )
        The variable for calculating the probability density.
    mu : float or 1-D array_like, (K, )
        The mean of the Gaussian distribution.
    cov : float or 2-D array_like, (K, K)
        The variance or the covariance matrix of the Gaussian distribution.
    log : bool
        If true, the return value is at log scale.

    Returns
    -------
    pdf : numpy float
        The probability density of x.
    """
    x = np.array(x) - np.array(mu)
    if len(np.array(cov).shape) < 2:
        cov = np.array(cov).reshape(-1,1)
    if len(np.array(x).shape) < 1:
        x = np.array(x).reshape(-1)
    cov_inv = np.linalg.inv(cov)
    cov_det = np.linalg.det(cov)
    if cov_det < 0:
        print("The det of covariance is negative, please check!")
        return None
    pdf = (-0.5*np.log(2*np.pi*cov_det) - 0.5*np.dot(np.dot(x, cov_inv), x))
    if log == False: pdf = np.exp(pdf)
    return pdf

def trun_normal_pdf(x, mu, sigma, a, b, log=True):
    """
    Calculate the probability density of Truncated Normal distribution.

    Parameters
    ----------
    x : float
        The variable for calculating the probability density.
    mu : float
        The mean of the Gaussian distribution.
    sigma : float
        The standard variance of the Gaussian distribution.
    a : float
        The lower bounder of the Truncated Normal distribution
    b : float
        The upper bounder of the Truncated Normal distribution
    log : bool
        If true, the return value is at log scale.

    Returns
    -------
    pdf : float
        The probability density of x.
    """
    x = x - mu
    a = a - mu
    b = b - mu
    
    pdf = np.exp(-0.5 * (x/sigma)**2) / (sigma * np.sqrt(2 * np.pi))
    cdf_a = (1 + erf(a / sigma / np.sqrt(2))) / 2.0
    cdf_b = (1 + erf(b / sigma / np.sqrt(2))) / 2.0
    
    pdf = pdf / abs(cdf_b - cdf_a)
    if log == True: pdf = np.log(pdf)
    return pdf

def GP_K(X, theta):
    """
    Covariance of Gaussian process generator.
    It is based on a common squared-exponential kernel, with two parameters. 

    Parameters
    ----------
    X : 1-D array_like, (N, )
        The x-axis of the Gaussian process, e.g., time points.
    theta : 1-D array_like, (2,)
        The array of the two parameters of the squared-exponential kernel.

    Returns
    -------
    K : 2-D array_like, (N, N)
        The covariance matrix of the N points at x-axis.
    """
    N = len(X)
    K = np.zeros((N, N))
    for i in range(N):
        for j in range(N):
            K[i,j] = theta[0] * np.exp(-0.5 * (X[i]-X[j])**2 / theta[1])
    return K

def Geweke_Z(X, first=0.1, last=0.5):
    """
    Geweke diagnostics for MCMC chain convergence.
    See Geweke J. Evaluating the accuracy of sampling-based approaches to the 
    calculation of posterior moments[M]. Minneapolis, MN, USA: Federal Reserve 
    Bank of Minneapolis, Research Department, 1991.
    and https://pymc-devs.github.io/pymc/modelchecking.html#formal-methods

    Parameters
    ----------
    X : 1-D array_like, (N, )
        The uni-variate MCMC sampled chain for convergence diagnostic.
    first : float
        The proportion of first part in Geweke diagnostics.
    last : float
        The proportion of last part in Geweke diagnostics.

    Returns
    -------
    Z : float
        The Z score of Geweke diagnostics.
    """
    N = X.shape[0]
    A = X[:int(first*N)]
    B = X[int(last*N):]
    if np.sqrt(np.var(A) + np.var(B)) == 0: 
        Z = None
    else:
        Z = abs(A.mean() - B.mean()) / np.sqrt(np.var(A) + np.var(B))
    return Z


def Psi_GP_MH(R_mat, len_isos, prob_isos, X=None, Ymean=None, var=None, 
              theta1=3.0, theta2=None, M=20000, initial=1000, gap=500, 
              randomS=None, theta2_std=1.0, theta2_low=0.01, theta2_up=100):
    """
    Estimate the proportion of C isoforms at T time points with all reads 
    by MCMC samplling (MH algorithm) combined with a GP prior.

    Parameters
    ----------
    R_mat : list of 2-D array_like, of length T
        A set of reads identities of belonging to C isoforms
    len_isos : list of 2-D array_like, of length T
        A set of effective length of C isoforms
    prob_isos : list of 2-D array_like, of length T
        A set of probablity for isoform specific reads
    X : 1-D array_like, (T, )
        An array of time points.
    Ymean : 2-D array_like, (C, T)
        The means for Y.
    var : float
        The average variation of all y.
    theta : 1-D array_like or list, (2, )
        The fixed hyper-parameter theta1, and initial theta2, default=[3.0,5.0]
    theta2_std : float
        The jump std of hyper-parameter theta2 for each dimension, default=1.0
    M : int
        the maximum iterations of in MCMC sampler, default=100000
    initial : int
        the minmum iterations of in MCMC sampler, default=3000
    gap : int
        the gap iterations of in MCMC sampler, default=1000

    Returns
    -------
    Psi_all : 3-D array_like, (m, C, T)
        The the proposed m proportion of C isoforms of T time points
     :
    Y_all : 3-D array_like, (m, C, T)
        The the proposed m latent y for C isoforms of T time points
    theta_all : 3-D array_like, (m, C-1, 2)
        The the proposed m hyper-parameters for C-1 isoforms
    Pro_all : 1-D array_like, (m,)
        The the probability for accepted proposals
    Lik_all : 1-D array_like, (m,)
        The the probability for accepted proposals
    cnt : int
        The number of acceptances
    m : int
        The number of iterations
    """
    T = len(len_isos)
    C = len(len_isos[0])
    if X is None: X = np.arange(T)
    if Ymean is None: Ymean = np.zeros((C,T))
    if randomS is not None: np.random.seed(randomS)
    for t in range(T):
        idx = (len_isos[t] != len_isos[t])
        len_isos[t][idx] = 0.0
        prob_isos[t][:,idx] = 0.0
        R_mat[t][:,idx] = False

        idx = np.where(R_mat[t] != R_mat[t])
        R_mat[t][idx] = False

        idx = np.where(prob_isos[t] != prob_isos[t])
        prob_isos[t][idx] = 0.0

        idx = (R_mat[t].sum(axis=1) > 0) * (prob_isos[t].sum(axis=1) > 0)
        R_mat[t] = R_mat[t][idx,:]
        prob_isos[t] = prob_isos[t][idx,:]

    # step 0: MCMC fixed initializations
    theta_now = np.zeros((C-1, 2))
    theta_now[:,0] = theta1
    theta_now[:,1] = theta2
    if var is None: var = 0.05 * np.ones(C-1)
    if theta2 is not None: theta_now[:,1] = theta2
    else: theta_now[:,1] = 0.75
    Y_now = Ymean + 0.0
    Ymean = np.zeros((C,T))

    psi_now = np.zeros((C, T))
    fsi_now = np.zeros((C, T))
    for t in range(T):
        psi_now[:,t] = np.exp(Y_now[:,t]) / np.sum(np.exp(Y_now[:,t]))
        fsi_now[:,t] = len_isos[t]*psi_now[:,t]/np.sum(len_isos[t]*psi_now[:,t])
    
    P_now, L_now = 0, 0
    cov_now = np.zeros((T, T, C-1))
    for c in range(C-1):
        cov_now[:,:,c] = GP_K(X, theta_now[c,:])
        P_now += normal_pdf(Y_now[c,:], Ymean[c,:], cov_now[:,:,c])
        
    for t in range(T):
        P_now += np.log(np.dot(R_mat[t]*prob_isos[t], fsi_now[:, t])).sum()
        L_now += np.log(np.dot(R_mat[t]*prob_isos[t], fsi_now[:, t])).sum()
        
    # MCMC running
    theta_try = np.zeros((C-1, 2))
    theta_all = np.zeros((M, C-1, 2))
    Y_try   = np.zeros((C, T))
    Y_all   = np.zeros((M, C, T))
    psi_try = np.zeros((C, T))
    fsi_try = np.zeros((C, T))
    Psi_all = np.zeros((M, C, T))
    cov_try = np.zeros((T, T, C-1))
    
    cnt = 0
    Pro_all = np.zeros(M)
    Lik_all = np.zeros(M)
    for m in range(M):
        P_try, L_try, Q_now, Q_try = 0, 0, 0, 0
        
        # step 1: propose a value
        for c in range(C-1):
            theta_try[c,0] = theta1
            if theta2 is not None: theta_try[c,1] = theta2
            else:
                theta_try[c,1] = np.random.normal(theta_now[c,1], theta2_std)
                while theta_try[c,1]<theta2_low or theta_try[c,1]>theta2_up:
                    theta_try[c,1] = np.random.normal(theta_now[c,1],theta2_std)
            cov_try[:,:,c] = GP_K(X, theta_try[c,:])
            Y_try[c,:] = np.random.multivariate_normal(Y_now[c,:], 
                cov_try[:,:,c]/theta1*var[c]*5/T/C)
                        
            P_try += normal_pdf(Y_try[c,:], Ymean[c,:], cov_try[:,:,c])
            Q_now += normal_pdf(Y_now[c,:], Y_try[c,:], 
                cov_try[:,:,c]/theta1*var[c]*5/T/C)
            Q_try += normal_pdf(Y_try[c,:], Y_now[c,:], 
                cov_now[:,:,c]/theta1*var[c]*5/T/C)
            Q_now += trun_normal_pdf(theta_now[c,1], theta_try[c,1], 
                theta2_std, theta2_low, theta2_up)
            Q_try += trun_normal_pdf(theta_try[c,1], theta_now[c,1], 
                theta2_std, theta2_low, theta2_up)
            
        for t in range(T):
            psi_try[:,t] = np.exp(Y_try[:,t]) / np.sum(np.exp(Y_try[:,t]))
            fsi_try[:,t] = (len_isos[t]*psi_try[:,t] / 
                np.sum(len_isos[t]*psi_try[:,t]))
            P_try += np.log(np.dot(R_mat[t]*prob_isos[t], fsi_try[:,t])).sum()
            L_try += np.log(np.dot(R_mat[t]*prob_isos[t], fsi_try[:,t])).sum()
        
        # step 2: calculate the MH ratio and accept or reject the proposal
        alpha = np.exp(min(P_try+Q_now-P_now-Q_try, 0))
        if alpha is None:
            print("alpha is none!")
        elif np.random.rand(1) < alpha:
            #print alpha
            cnt += 1
            P_now = P_try + 0.0
            L_now = L_try + 0.0
            Y_now = Y_try + 0.0
            cov_now = cov_try + 0.0
            psi_now = psi_try + 0.0
            fsi_now = fsi_try + 0.0
            theta_now = theta_try + 0.0            

        Pro_all[m]     = P_now
        Lik_all[m]     = L_now
        Y_all[m,:,:]   = Y_now
        Psi_all[m,:,:] = psi_now
        theta_all[m,:] = theta_now
        
        #step 3. convergence diagnostics
        if m >= initial and m % gap == 0:
            conv = 1
            for c in range(C-1):
                for t in range(T):
                    # Z = Geweke_Z(Y_all[:m,c,t])
                    Z = Geweke_Z(Psi_all[:m,c,t])
                    if Z is None or Z > 2: 
                        conv = 0
                        break
                Z = Geweke_Z(theta_all[:m,c,1])
                if Z is None or Z > 2: conv == 0
                if conv == 0: break
            if conv == 1:
                Pro_all = Pro_all[:m,]
                Lik_all = Lik_all[:m,]
                Y_all   = Y_all[:m,:,:]
                Psi_all = Psi_all[:m,:,:]
                theta_all = theta_all[:m,:]
                break
    return Psi_all, Y_all, theta_all, Pro_all, Lik_all, cnt, m


def EM_filter(R_mat, len_isos, prob_isos, min_num=2):
    T = len(len_isos)
    C = len(len_isos[0])
    for t in range(T):
        idx = (len_isos[t] != len_isos[t])
        len_isos[t][idx] = 0.0
        prob_isos[t][:,idx] = 0.0
        R_mat[t][:,idx] = False

        idx = np.where(R_mat[t] != R_mat[t])
        R_mat[t][idx] = False

        idx = np.where(prob_isos[t] != prob_isos[t])
        prob_isos[t][idx] = 0.0

        idx = (R_mat[t].sum(axis=1) > 0) * (prob_isos[t].sum(axis=1) > 0)
        R_mat[t] = R_mat[t][idx,:]
        prob_isos[t] = prob_isos[t][idx,:]

    KP_FLAG = np.ones(C, 'bool')
    count_tmp = np.zeros(T)
    psi_tmp = np.zeros((T, C))
    for t in range(T):
        if np.sum(R_mat[t]) == 0: continue
        count_tmp = np.sum(R_mat[t][:,:].sum(axis=1) > 0)
        psi_tmp[t,:], _Lik_all = Psi_EM(R_mat[t], len_isos[t], prob_isos[t])

    idx = np.argsort(psi_tmp.mean(axis=0))
    for i in range(len(idx)-min_num):
        if np.max(psi_tmp[:,idx[i]]) < 0.001 and np.mean(count_tmp) > 100:
            KP_FLAG[idx[i]] = False
    return KP_FLAG, psi_tmp


def Psi2Y(Psi):
    Psi = Psi / np.sum(Psi)
    Y = np.zeros(len(Psi))
    Y[:-1] = np.log(Psi[:-1] / Psi[-1])
    return Y

def EM_bootstrap(R_mat, len_isos, prob_isos, bootN=10):
    T = len(len_isos)
    C = len(len_isos[0])
    for t in range(T):
        idx = (len_isos[t] != len_isos[t])
        len_isos[t][idx] = 0.0
        prob_isos[t][:,idx] = 0.0
        R_mat[t][:,idx] = False

        idx = np.where(R_mat[t] != R_mat[t])
        R_mat[t][idx] = False

        idx = np.where(prob_isos[t] != prob_isos[t])
        prob_isos[t][idx] = 0.0

        idx = (R_mat[t].sum(axis=1) > 0) * (prob_isos[t].sum(axis=1) > 0)
        R_mat[t] = R_mat[t][idx,:]
        prob_isos[t] = prob_isos[t][idx,:]

    Y_boot = np.zeros((bootN, T, C))
    psi_boot = np.ones((bootN, T, C))/(C+0.0)
    for i in range(bootN):
        for t in range(T):
            _N = R_mat[t].shape[0]
            if _N == 0: continue
            idx = np.random.randint(_N, size=_N)
            if np.sum(R_mat[t][idx,:]) == 0: continue
            psi_boot[i,t,:], _Lik_all = Psi_EM(R_mat[t][idx,:], len_isos[t], prob_isos[t][idx,:])
            Y_boot[i,t:] = Psi2Y(psi_boot[i,t,:] + 0.000001)
    var_boot = Y_boot.var(axis=0).mean(axis=0)
    # var_boot = psi_boot.var(axis=0).mean(axis=0) * 36
    
    return Y_boot.mean(axis=0), var_boot
    # return psi_boot.mean(axis=0), var_boot

def Psi_EM(R_mat, len_isos, prob_isos, M=100):
    # check input data
    idx = (len_isos != len_isos)
    len_isos[idx] = 0.0
    prob_isos[:,idx] = 0.0
    R_mat[:,idx] = False

    idx = np.where(R_mat != R_mat)
    R_mat[idx] = False

    idx = np.where(prob_isos != prob_isos)
    prob_isos[idx] = 0.0

    idx = (R_mat.sum(axis=1) > 0) * (prob_isos.sum(axis=1) > 0)
    R_mat = R_mat[idx,:]
    prob_isos = prob_isos[idx,:]

    # random initialization
    Psi = np.random.dirichlet(np.ones(R_mat.shape[1]))
    Fsi = Psi * len_isos / np.sum(Psi * len_isos)
    Lik = np.log(np.dot(R_mat*prob_isos, Fsi)).sum()
    
    cnt = 0
    Lik_all = np.zeros(M)
    for i in range(M):
        cnt += 1
        # E step
        Num = (R_mat*prob_isos*Fsi)
        Num = Num / Num.sum(axis=1).reshape(-1,1)
        Num = Num.sum(axis=0)
        
        # M step
        Pro = Num / len_isos
        Psi = Pro / sum(Pro)
        Fsi = Psi * len_isos / sum(Psi * len_isos)
        Lik = np.log(np.dot(R_mat*prob_isos, Fsi)).sum()
        Lik_all[i] = Lik

        # check convergence
        if i > 0 and (Lik - Lik_all[i-1]) < np.log(1+10**(-5)):
            break
    return Psi, Lik_all


# Copyright(c) 2014, The DICEseq developer (Yuanhua Huang)
# Licensed under the MIT License at
# http://opensource.org/licenses/MIT

__version__ = "0.1.6"

from .utils.reads_utils import ReadSet
from .utils.out_utils import DiceFile, SampleFile
from .utils.bias_utils import FastaFile, BiasFile
from .utils.tran_utils import TranUnits, TranSplice
from .utils.sam_utils import load_samfile, fetch_reads
from .utils.gtf_utils import Gene, Transcript, load_annotation

from .models.model_GP import Psi_GP_MH
from .models.model_static import Psi_MCMC_MH, Psi_analytic, Psi_junction




# This file is to merge bias parameters from all bias files.

import h5py
import numpy as np
from optparse import OptionParser
from .utils.bias_utils import BiasFile

def main():
   #part 0. parse command line options
    parser = OptionParser()
    parser.add_option("--bias_list", dest="bias_list", default=None,
        help="The list of bias file path, separated by ---")
    parser.add_option("--out_file", dest="out_file", default=None,
        help="The merged bias file for output.")

    #part 0.1. load the data
    (options, args) = parser.parse_args()
    file_list = list_file.split("---")
    out_file  = options.out_file

    #part 1.1. copy a file as output file
    N = len(file_list)
    biasFile  = BiasFile(file_list[0])
    biasFile.pos5_bias /= N
    biasFile.pos3_bias /= N
    biasFile.pos5_unif /= N
    biasFile.pos3_unif /= N
    for i in range(len(biasFile.chain_len)):
        biasFile.seq5_bias[str(i)] /= N
        biasFile.seq3_bias[str(i)] /= N
        biasFile.seq5_unif[str(i)] /= N
        biasFile.seq3_unif[str(i)] /= N

    #part 1.2. merge other files
    for n in np.arange(1, N):
        _biasFile = BiasFile(file_list[n])

        biasFile.pos5_bias += _biasFile.pos5_bias / N
        biasFile.pos3_bias += _biasFile.pos3_bias / N
        biasFile.pos5_unif += _biasFile.pos5_unif / N
        biasFile.pos3_unif += _biasFile.pos3_unif / N
        for i in range(len(biasFile.chain_len)):
            biasFile.seq5_bias[str(i)] += _biasFile.seq5_bias[str(i)] / N
            biasFile.seq3_bias[str(i)] += _biasFile.seq3_bias[str(i)] / N
            biasFile.seq5_unif[str(i)] += _biasFile.seq5_unif[str(i)] / N
            biasFile.seq3_unif[str(i)] += _biasFile.seq3_unif[str(i)] / N

    #part 2. save file
    biasFile.save_file(out_file) 

if __name__ == "__main__":
    main()



# This module is to parse the gtf file, whoes format can be found at:
# www.ensembl.org/info/website/upload/gff.html

# Note: here we need the order of feature lines like this: gene --> transcript
# --> exon 1 --> exon 2 ..., which is not required for the gtf format. But this
# assumption is usually right for the gtf file downloaded from Ensembl database.

import sys
import numpy as np

class Transcript:
    def __init__(self,chrom,strand,start,stop,tran_id,tran_name="*",biotype="*"):
        """a general purpose transcript object with the basic information.
        """
        self.chrom  = chrom
        self.strand = strand
        self.start  = int(start)
        self.stop   = int(stop)
        self.tranID = tran_id
        self.exons  = np.zeros((0,2), "int")
        self.seglen = None
        self.tranL  = 0
        self.exonNum = 0
        self.biotype = biotype
        self.tranName = tran_name
        

    def add_exon(self,chrom,strand,start,stop):
        if strand != self.strand or chrom != self.chrom:
            print("The exon has different chrom or strand to the transcript.")
            return
        _exon = np.array([start, stop], "int").reshape(1,2)
        self.exons = np.append(self.exons, _exon, axis=0)
        self.exons = np.sort(self.exons, axis=0)
        self.tranL += abs(int(stop) - int(start) + 1)
        self.exonNum += 1


        self.seglen = np.zeros(self.exons.shape[0] * 2 - 1, "int")
        self.seglen[0] = self.exons[0,1]-self.exons[0,0] + 1
        for i in range(1, self.exons.shape[0]):
            self.seglen[i*2-1] = self.exons[i,0]-self.exons[i-1,1] - 1
            self.seglen[i*2] = self.exons[i,1]-self.exons[i,0] + 1

        if ["-","-1","0",0,-1].count(self.strand) > 0:
            self.seglen = self.seglen[::-1]

class Gene:
    def __init__(self,chrom,strand,start,stop,gene_id,gene_name="*",biotype="*"):
        self.chrom  = chrom
        self.strand = strand
        self.start  = int(start)
        self.stop   = int(stop)
        self.geneID = gene_id
        self.trans  = []
        self.tranNum = 0
        self.biotype = biotype
        self.geneName = gene_name
        
    def add_transcipt(self, transcript):
        self.trans.append(transcript)
        self.tranNum += 1

    def get_gene_info(self):
        RV = [self.geneID, self.geneName, self.chrom, self.strand, self.start,
              self.stop, self.biotype]
        _trans = []
        for t in self.trans:
            _trans.append(t.tranID)
        RV.append(",".join(_trans))
        return RV

    def add_premRNA(self):
        _tran = Transcript(self.chrom, self.strand, self.start, self.stop, 
                           self.geneID+".p", self.geneName, self.biotype)
        _tran.add_exon(self.chrom, self.strand, self.start, self.stop)
        self.trans.append(_tran)
        self.tranNum += 1
        
    def get_exon_max_num(self):
        exonMax = 0
        for _tran in self.trans:
            exonMax = max(exonMax, _tran.exonNum)
        return exonMax

    def gene_ends_update(self):
        for t in self.trans:
            self.start = min(self.start, np.min(t.exons))
            self.stop  = max(self.stop,  np.max(t.exons))

def ensembl_gtf(anno_in):
    genes = []
    _gene = None
    for _line in anno_in:
        # comment lines
        if _line[0] == "#" or _line[0] == ">" : continue
            
        a_line = _line.split("\t")
        if len(a_line) < 9: continue
        elif a_line[2] == "gene":
            if _gene is not None:
                genes.append(_gene)
                _gene = None

            _gene_name, _gene_id, _biotype = "*", "*", "*"
            idx = a_line[8].find("gene_name")
            if idx > -1:
                _gene_name = a_line[8][idx:].split('"')[1].split("\n")[0]
            idx  = a_line[8].find("gene_id")
            if idx > -1:
                _gene_id = a_line[8][idx:].split('"')[1].split("\n")[0]
            idx = a_line[8].find("gene_biotype")
            if idx == -1: idx = a_line[8].find("gene_type")
            if idx > -1:
                _biotype = a_line[8][idx:].split('"')[1].split("\n")[0]

            _gene = Gene(a_line[0], a_line[6], a_line[3], a_line[4],
                         _gene_id, _gene_name, _biotype)

        elif a_line[2] == "transcript":
            _tran_name, _tran_id, _biotype = "*", "*", "*"

            idx = a_line[8].find("transcript_name")
            if idx > -1:
                _tran_name = a_line[8][idx:].split('"')[1].split("\n")[0]
            idx = a_line[8].find("transcript_id")
            if idx > -1:
                _tran_id = a_line[8][idx:].split('"')[1].split("\n")[0]
            idx = a_line[8].find("gene_biotype")
            if idx > -1:
                _biotype = a_line[8][idx:].split('"')[1]
            _tran = Transcript(a_line[0], a_line[6], a_line[3], a_line[4],
                               _tran_id, _tran_name, _biotype)

            if _gene is not None:
                _gene.add_transcipt(_tran)
            else:
                print("Gene is not ready before transcript.")

        elif a_line[2] == "exon":
            if a_line[0] != _gene.trans[-1].chrom:
                print("The exon is on the different chrom from the transcript.")
                continue
            if a_line[6] != _gene.trans[-1].strand:
                print("The exon is on the different strand from the transcript.")
                continue
            if _gene is not None and len(_gene.trans) > 0:
                _gene.trans[-1].add_exon(a_line[0],a_line[6],a_line[3],a_line[4])
                # _gene.gene_ends_update()
            else:
                print("Gene or transcript is not ready before exon.")

    if _gene is not None: genes.append(_gene)
    return genes


def sander_gtf(anno_in):
    genes = []
    _gene = None
    for _line in anno_in:
        # comment lines
        if _line[0] == "#" or _line[0] == ">" : continue
            
        a_line = _line.split("\t")
        if len(a_line) < 9: continue
        elif a_line[2] == "exon":
            _gene_id = "#"
            idx  = a_line[8].find("gene_id")
            if idx > -1:
                _gene_id = a_line[8][idx:].split('"')[1]
            if len(genes) > 0 and _gene_id == genes[-1].trans[-1].tranID :
                genes[-1].trans[-1].add_exon(a_line[0], a_line[6],
                                             a_line[3], a_line[4])
                genes[-1].gene_ends_update()
            else:
                _biotype = a_line[1]
                _gene_name, _gene_id = "*", "*"
                idx = a_line[8].find("gene_name")
                if idx > -1:
                    _gene_name = a_line[8][idx:].split('"')[1].split("\n")[0]
                idx  = a_line[8].find("gene_id")
                if idx > -1:
                    _gene_id = a_line[8][idx:].split('"')[1].split("\n")[0]
                _gene = Gene(a_line[0], a_line[6], a_line[3], a_line[4],
                             _gene_id, _gene_name, _biotype)

                _tran_name, _tran_id = "*", "*"
                idx = a_line[8].find("transcript_name")
                if idx > -1:
                    _tran_name = a_line[8][idx:].split('"')[1].split("\n")[0]
                idx = a_line[8].find("transcript_id")
                if idx > -1:
                    _tran_id = a_line[8][idx:].split('"')[1].split("\n")[0]
                idx = a_line[8].find("gene_biotype")
                _tran = Transcript(a_line[0], a_line[6], a_line[3], a_line[4],
                                   _tran_id, _tran_name, _biotype)

                _tran.add_exon(a_line[0], a_line[6], a_line[3], a_line[4])

                _gene.add_transcipt(_tran)
                genes.append(_gene)
    return genes


def sgd_gtf(anno_in):
    genes = []
    _gene = None
    cnt =+ 0
    for _line in anno_in:
        # comment lines
        if _line[0] == "#":continue
        elif _line[0] == ">" : break
            
        a_line = _line.split("\t")
        if len(a_line) < 9: continue
        elif a_line[2].find("gene") > -1:
            if _gene is not None:
                genes.append(_gene)
                _gene = None
            
            if a_line[2].find("gene") == 0:
                _biotype = "protein_coding"
            else:
                _biotype = a_line[2].split("_")[0]
            _gene_name, _gene_id = "*", "*"
            idx = a_line[8].find("Name")
            if idx > -1:
                _gene_name = a_line[8][idx:].split(";")[0].split("=")[1].split("\n")[0]
            idx  = a_line[8].find("ID")
            if idx > -1:
                _gene_id = a_line[8][idx:].split(";")[0].split("=")[1].split("\n")[0]
            _gene = Gene(a_line[0], a_line[6], a_line[3], a_line[4],
                         _gene_id, _gene_name, _biotype)

        # elif a_line[2] == "mRNA":
        #     _tran_name, _tran_id, _biotype = "*", "*", "*"
        #     idx = a_line[8].find("Name")
        #     if idx > -1:
        #         _tran_name = a_line[8][idx:].split(";")[0].split("=")[1]
        #     idx = a_line[8].find("ID")
        #     if idx > -1:
        #         _tran_id = a_line[8][idx:].split(";")[0].split("=")[1]
            # _tran = Transcript(a_line[0], a_line[6], a_line[3], a_line[4],
            #                    _tran_id, _tran_name, _biotype)

            _tran_name, _tran_id = _gene_name, _gene_id
            _tran = Transcript(a_line[0], a_line[6], a_line[3], a_line[4],
                               _tran_id, _tran_name, _biotype)

            if _gene is not None:
                _gene.add_transcipt(_tran)
            else:
                print("Gene is not ready before transcript.")

        elif a_line[2] == "CDS" or a_line[2] == "noncoding_exon" :
            if _gene is not None and len(_gene.trans) > 0:
                _gene.trans[-1].add_exon(a_line[0],a_line[6],a_line[3],a_line[4])
                # _gene.gene_ends_update()
            else:
                print("Gene or transcript is not ready before exon.")

    if _gene is not None: genes.append(_gene)
    return genes


def miso_gtf(anno_in):
    genes = []
    _gene = None
    for _line in anno_in:
        # comment lines
        if _line[0] == "#" or _line[0] == ">" : continue
            
        a_line = _line.split("\t")
        if len(a_line) < 9: continue
        elif a_line[2] == "gene":
            if _gene is not None:
                genes.append(_gene)
                _gene = None

            _gene_name, _gene_id, _biotype = "*", "*", "*"
            idx = a_line[8].find("Name")
            if idx > -1:
                _gene_name = a_line[8][idx:].split(";")[0].split("=")[1].split("\n")[0]
            idx  = a_line[8].find("ID=")
            if idx > -1:
                _gene_id = a_line[8][idx:].split(";")[0].split("=")[1].split("\n")[0]
            _gene = Gene(a_line[0], a_line[6], a_line[3], a_line[4],
                         _gene_id, _gene_name, _biotype)

        elif a_line[2] == "mRNA":
            _tran_name, _tran_id, _biotype = "*", "*", "*"
            idx = a_line[8].find("ID=")
            if idx > -1:
                _tran_name = a_line[8][idx:].split(";")[0].split("=")[1].split("\n")[0]       
            idx = a_line[8].find("ID=")
            if idx > -1:
                # print(a_line[8][idx:])
                _tran_id = a_line[8][idx:].split(";")[0].split("=")[1].split("\n")[0]
            _tran = Transcript(a_line[0], a_line[6], a_line[3], a_line[4],
                               _tran_id, _tran_name, _biotype)

            if _gene is not None:
                _gene.add_transcipt(_tran)
            else:
                print("Gene is not ready before transcript.")

        elif a_line[2] == "exon":
            if _gene is not None and len(_gene.trans) > 0:
                _gene.trans[-1].add_exon(a_line[0],a_line[6],a_line[3],a_line[4])
                # _gene.gene_ends_update()
            else:
                print("Gene or transcript is not ready before exon.")
    
    if _gene is not None: genes.append(_gene)
    return genes

def load_annotation(anno_gtf, source="Ensembl"):
    """load gene annotation in convenient structures;
    we support several source of gtf files: Ensembl, SGD, MISO, Sander.
    """
    # Note: we need the start and stop site of genes but only transcripts,
    # especially there are multiple transcripts on a gene.

    fid = open(anno_gtf, "r")
    anno_in = fid.readlines()
    fid.close()
    
    gene_name, gene_id, genes = [], [], []
    strand, exon_num, biotype = [], [], []
    chrom, gene_start, gene_stop = [], [], []
    tran_num = []

    if source == "SGD" or source == "sgd":
        genes = sgd_gtf(anno_in)
    elif source == "MISO" or source == "miso":
        genes = miso_gtf(anno_in)
    elif source == "Sander" or source == "sander":
        genes = sander_gtf(anno_in)
    else:
        genes = ensembl_gtf(anno_in)

    for g in genes:
        exon_num.append(g.get_exon_max_num())
        chrom.append(g.chrom)
        strand.append(g.strand)
        gene_start.append(g.start)
        gene_stop.append(g.stop)
        gene_id.append(g.geneID)
        gene_name.append(g.geneName)
        biotype.append(g.biotype)
        tran_num.append(g.tranNum)

    RV = {}
    RV['genes'] = genes
    RV['chrom'] = np.array(chrom)
    RV['strand'] = np.array(strand)
    RV['gene_id'] = np.array(gene_id)
    RV['biotype'] = np.array(biotype)
    RV['exon_num'] = np.array(exon_num)
    RV['tran_num'] = np.array(tran_num)
    RV['gene_name'] = np.array(gene_name)
    RV['gene_stop'] = np.array(gene_stop)
    RV['gene_start'] = np.array(gene_start)
    return RV
    



# This is a main file to run the isplice software, which will return the 
# splicing ratio for each gene.

import sys
import h5py
import pysam
import numpy as np
from .optparse import OptionParser
from .models.model_static import Psi_MCMC_MH
from .utils.gtf_utils import load_annotation
from .utils.bias_utils import BiasFile, FastaFile
from .utils.sam_utils import load_samfile, fetch_reads
from .utils.tran_utils import TranUnits, Transcript, TranSplice

def get_CI(data, percent=0.95):
    if len(data.shape) == 0:
        data = data.reshape(-1,1)
    for k in range(data.shape[1]):
        data[:,k] = np.sort(data[:,k])
        CI_idx = int(data.shape[0] * (1-percent)/2)
        RV = data[-CI_idx,:] - data[CI_idx,:]
    return RV

def main():
    #part 0. parse command line options
    parser = OptionParser()
    parser.add_option("--anno_file", "-a", dest="anno_file", default=None,
        help="The annotation file in gtf format")
    parser.add_option("--sam_file", "-s", dest="sam_file", default=None,
        help="The indexed alignement file in bam/sam format")
    parser.add_option("--ref_file", "-r", dest="ref_file", default=None,
        help="The genome reference file in fasta format")
    parser.add_option("--bias_file", "-b", dest="bias_file", default=None,
        help="The parameter file for bias in hdf5 format")
    parser.add_option("--bias_mode", dest="bias_mode", default="unif", 
        help="The bias mode")
    parser.add_option("--gene_file", "-g", dest="gene_file", default=None,
        help="The list genes in use.")
    parser.add_option("--out_file", "-o", dest="out_file",
        default="estimated_rate.psi", help="The output file in txt format")

    (options, args) = parser.parse_args()
    if options.anno_file == None:
        print("Error: need --anno_file for annotation.")
        sys.exit(1)
    else:
        anno = load_annotation(options.anno_file)
    if options.sam_file == None:
        print("Error: need --sam_file for reads indexed and aliged reads.")
        sys.exit(1)
    else:
        sam_list = options.sam_file.split("---")
        samFiles = []
        for s in sam_list:
            samFiles.append(load_samfile(s))

    if options.gene_file == None:
        gene_list = anno["gene_id"]
    else:
        gene_list = np.loadtxt(options.gene_file, dtype="str")
    
    ref_file  = options.ref_file
    out_file  = options.out_file
    bias_file = options.bias_file
    bias_mode = options.bias_mode

    if ref_file is None: fastaFile = None
    else: fastaFile = FastaFile(ref_file)

    if bias_file is None: biasFile = None
    else: biasFile = BiasFile(bias_file)

    print("Welcome to diceseq static model! %d genes are under estimating." %len(gene_list))

    fid = open(out_file, "w")
    head_line = "gene_id\tgene_name\t"
    head_line = head_line + "count\tpsi_iso1\tpsi_iso2\tCI95_iso1\tCI95_iso2\n"
    fid.writelines(head_line)
    
    # 1 set the parameters for Psi inference MCMC method
    M = 100000
    cov_fixed = np.array([[0.15]])
    alpha_diri = np.ones(2) / 2

    # 2. run the model
    Psi_all = np.zeros((gene_list.shape[0],2))
    CIs_all = np.zeros((gene_list.shape[0],2))
    for g in range(gene_list.shape[0]):
        i = np.where(np.array(anno["gene_id"]) == gene_list[g])[0][0]
        Psi_all[g,:] = None

        _gene   = anno["gene_id"][i]
        # _exons  = anno["exons"][i]
        _strand = anno["strand"][i]
        _chrom  = str(anno["chrom"][i])
        _start  = anno["gene_start"][i]
        _stop   = anno["gene_stop"][i]
        # print (g, _gene, _chrom, _start, _stop, _strand)

        # t = TranSplice(_gene, _chrom, _strand, _start, _stop, _exons)
        # t.initial_unitSet()
        t = TranSplice(anno["genes"][i])
        t.initial_unitSet(add_premRNA=True)
        for samFile in samFiles:
            t.set_reads(samFile)

        if bias_mode != "unif":
            if fastaFile is None:
                print ("we re-set bias mode as uniform as reference " +
                "sequence are not available.")
                bias_mode = "unif"
            elif biasFile is None:
                print ("we re-set bias mode as uniform as bias " +
                "parameters are not available.")
                bias_mode = "unif"
            else:
                t.set_sequence(fastaFile)
                t.set_bias(biasFile)

        t.get_ready(bias_mode)
        count = len(t.read1p) + len(t.read1u) + len(t.read2u)
        if count == 0:
            print("No reads for %s" %anno["gene_id"][i])
            tmp = []; tmp[:] = ["nan"] * 5; tmp[0] = "0"
            a_line = anno["gene_id"][i] + "\t" + anno["gene_name"][i]
            a_line = a_line + "\t" + "\t".join(tmp) + "\n"
            fid.writelines(a_line)
            continue

        elif bias_mode == "unif":
            Psi_tmp = Psi_MCMC_MH(t.Rmat, t.efflen_unif, t.proU, cov_fixed,
                alpha_diri, M)
        else:
            Psi_tmp = Psi_MCMC_MH(t.Rmat, t.efflen_bias, t.proB, cov_fixed,
                alpha_diri, M)

        Psi_all[g,:] = np.mean(Psi_tmp[Psi_tmp.shape[0]/2:,:], axis=0)
        CIs_all[g,:] = get_CI(Psi_tmp[Psi_tmp.shape[0]/2:,:])

        # 3 save output
        rInfo = np.append(Psi_all[g,:], CIs_all[g,:])
        rInfo = list(np.append(count, rInfo))
        aLine = anno["gene_id"][i] + "\t" + anno["gene_name"][i]
        aLine = aLine + "\t" + "\t".join(["%.3f" %x for x in rInfo]) + "\n"
        fid.writelines(aLine)
    fid.close()

if __name__ == "__main__":
    main()
    



# This module is to parse the sam/bam file with the pysam tool, 
# which can be found at: http://pysam.readthedocs.org

import sys
import pysam

def load_samfile(sam_file):
    """To load an indexed bam file"""
    ftype = sam_file.split(".")[-1]
    if ftype != "bam" and ftype != "sam":
        print("Error: file type need suffix of bam or sam.")
        sys.exit(1)
    print("Loading a %s" %ftype + " with file name: %s" %sam_file)
    if ftype == "bam":
        samfile = pysam.Samfile(sam_file, "rb")
    else:
        samfile = pysam.Samfile(sam_file, "r")
    return samfile


def fetch_reads(samfile, chrom, start, end, rm_duplicate=True, inner_only=True,
                mapq_min=0, mismatch_max=10**5, rlen_min=1,  is_mated=True):
    """To fetch the reads in a given region from a pysam AlignmentFile.

    Args:
        samfile: A Samfile object in pysam.
        chrom: A string of chromosome, e.g., "IV", "chr10".
        start: An integer of the start position for mapped reads.
        end: An integer of the end position for mapped reads.
        rm_duplicate: A bool for only keeping the first one of duplicates.
        inner_only: A bool for only keeping fully region matched reads.
        mapq_min: An integer of the minimum of map quality.
        mismatch_min: An integer of the minimum of mismatched bases.
        rlen_min: An integer of the minimum of read length.
        is_paired: A bool for mating paired-end reads.

    Returns:
        A dict containing lists of mated reads1 and reads2, and unmated reads1u
        and reads2u, i.e.,
        {'reads1': [r11, r21, ...]
         'reads2': [r12, r22, ...]
         'reads1u': [r*1, r*1, ...]
         'reads2u': [r*2, r*2, ...]}
        reads1 is the 5-end of the fragment, and reads2 is the 3-end of the 
        fragment.

    Raises:
        ValueError: An error occurred when fetching reads.
        AssertionError: An error occurred when fetching reads.
    """
    #part 1. check the input and fetch the reads
    chrom  = str(chrom)
    if chrom in samfile.references:
        pass
    else:
        chrom_parts = chrom.split("chr")
        if len(chrom_parts) <= 1:
            chrom = chrom_parts[0]
        else:
            chrom = chrom_parts[1]
    try:
        reads = samfile.fetch(chrom, start, end)

    except ValueError:
        print("Cannot fetch reads in region: %s:%d-%d" %(chrom, start, end))
    except AssertionError:
        print("AssertionError in region: %s:%d-%d" %(chrom, start, end))
        print(" - Check that your BAM file is indexed!")

    #part 2. get reads and filter some of them
    qname1, qname2 = [], []
    reads1, reads2 = [], []
    r_prev = None
    for r in reads:
        # filter 4: only keep the first one of duplicates
        if (rm_duplicate and r_prev is not None and r_prev.qname == r.qname and 
            r_prev.positions == r.positions): r_prev = r; continue
        r_prev = r
        # filter 1: only particially mapped to the regions
        if inner_only == True and (r.pos < start or r.aend > end): continue
        # filter 2: too low map quality
        if r.mapq < mapq_min: continue
        # filter 3: too long mismatch
        if r.rlen - len(r.positions) > mismatch_max: continue
        # filter 5: too short reads length
        if r.rlen < rlen_min: continue
        
        if r.is_read2:
            reads2.append(r)
            qname2.append(r.qname)
        else:
            reads1.append(r)
            qname1.append(r.qname)

    #part 2.1 chech the mate reads' query
    FLAG = True
    if len(qname1) > 0:
        for i in range(len(qname1)-1):
            if qname1[i][-1] != qname1[i+1][-1]:
                FLAG = False
                break
    if FLAG and len(qname2) > 0:
        for i in range(len(qname1)-1):
            if qname1[i][-1] != qname1[i+1][-1]:
                FLAG = False
                break

    if FLAG:
        for i in range(len(qname1)):
            qname1[i] = qname1[i][:-1]
        for i in range(len(qname2)):
            qname2[i] = qname2[i][:-1]


    # part 3. mate the reads
    rv_reads1, rv_reads2 = [], []
    rv_reads1u, rv_reads2u = [], []
    if is_mated == True:
        idx1 = sorted(range(len(qname1)), key=qname1.__getitem__)
        idx2 = sorted(range(len(qname2)), key=qname2.__getitem__)

        i1, i2 = 0, 0
        while i1 < len(idx1) and i2 < len(idx2):
            if qname1[idx1[i1]] == qname2[idx2[i2]]:
                rv_reads1.append(reads1[idx1[i1]])
                rv_reads2.append(reads2[idx2[i2]])
                i1, i2 = i1 + 1, i2 + 1
            elif qname1[idx1[i1]] < qname2[idx2[i2]]:
                rv_reads1u.append(reads1[idx1[i1]])
                i1 += 1
            elif qname1[idx1[i1]] > qname2[idx2[i2]]:
                rv_reads2u.append(reads2[idx2[i2]])
                i2 += 1
        for i in range(i1, len(idx1)):
            rv_reads1u.append(reads1[idx1[i]])
        for i in range(i2, len(idx2)):
            rv_reads2u.append(reads2[idx2[i]])
    else:
        rv_reads1u, rv_reads2u = reads1, reads2

    # part 4. return reads
    RV = {}
    RV["reads1"] = rv_reads1
    RV["reads2"] = rv_reads2
    RV["reads1u"] = rv_reads1u
    RV["reads2u"] = rv_reads2u
    return RV
    



"""
DICEseq - Dynamic Isoform spliCing Estimator via sequencing data
See: http://diceseq.sourceforge.net
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()
    
reqs = ['numpy', 'pysam', 'matplotlib']

setup(
    name='diceseq',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.1.7',

    description='DICEseq: Dynamic Isoform spliCing Estimator via sequencing data',
    long_description=long_description,

    # The project's main homepage.
    url='http://diceseq.sourceforge.net',

    # Author details
    author='Yuanhua Huang',
    author_email='Y.Huang@ed.ac.uk',

    # Choose your license
    license='MIT',

    # What does your project relate to?
    keywords=['splicing isoform quantification', 'time series RNA-seq',
              'Gaussian process', 'Markov chain Monte Carlo'],

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(),#['diceseq', 'diceseq.utils', 'diceseq.models'],#

    entry_points={
          'console_scripts': [
              'diceseq = diceseq.diceseq:main',
              'dice-count = diceseq.dice_count:main',
              'dice-bias = diceseq.bias_estimate:main',
              ],
          }, 

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    
    install_requires=reqs,

    py_modules = ['diceseq']

    # buid the distribution: python setup.py sdist
    # upload to pypi: twine upload dist/...

)


# This module is an interface of the diceseq output,
# with which you could easily access to the results,
# also visualize the data.

import gzip
import numpy as np
import pylab as pl

class SampleFile:
    """docstring for SampleFile"""
    def __init__(self, sample_file):
        fid = gzip.open(sample_file, "r")
        all_lines = fid.readlines()
        fid.close()
        sample = []
        self.gene, self.tran = [], []
        self.theta, self.sample = [], []
        for line in all_lines:
            line = line.split("\n")[0]
            if len(self.gene) == 0 and line[0] != "@": continue
            if line.startswith("@"):
                self.gene.append(line[1:].split("|")[0])
                self.tran.append(line[1:].split("|")[1].split(","))
                self.theta.append(np.array(line[1:].split("|")[2].split(","),"float"))
                if sample == []: continue
                self.sample.append(np.array(sample, "float"))
                sample = []
            else:
                _sample = line.split(";")
                T = len(line.split(";"))
                for i in range(len(_sample)):
                    _sample[i] = _sample[i].split(",")
                sample.append(_sample)
        self.sample.append(np.array(sample, "float"))

    def GP_visual(self, gquery, T, xx, thetas=None):
        """visualization function for Gaussian process"""
        if self.gene.count(gquery) == 0:
            print("No %s in the sample file." %gquery)
            exit(1)
        else: qidx = self.gene.index(gquery)
        _tran = self.tran[qidx]
        _sample = self.sample[qidx]
        
        yy = np.zeros((_sample.shape[0], xx.shape[0], len(_tran)))
        for c in range(len(_tran)):
            if c < len(_tran)-1:
                if thetas is None: 
                    theta1 = 3.0
                    theta2 = self.theta[qidx][c]
                    Theta = [theta1, theta2]
                elif len(np.array(thetas).shape) == 1:
                    Theta = thetas
                else:
                    Theta = thetas[c]    
            yy[:,:,c] = GP_sample_predict(Theta, T, _sample[:,:,c], xx)

        psi_mean = YtoPsi(_sample, axis=2).mean(axis=0)
        psi_pred = YtoPsi(yy, axis=2).mean(axis=0)
        psi_CI95 = np.zeros((len(T), len(_tran), 2))
        psi_pred95 = np.zeros((xx.shape[0], len(_tran), 2))
        for t in range(len(T)):
            psi_CI95[t,:,:] = get_CI(YtoPsi(_sample[:,t,:], axis=1))
        for t in range(xx.shape[0]):
            psi_pred95[t,:,:] = get_CI(YtoPsi(yy[:,t,:], axis=1))

        np.random.seed(10)
        for c in range(len(_tran)):
            _color = np.random.rand(3,1)
            psi_err = [psi_mean[:,c]-psi_CI95[:,c,1], psi_CI95[:,c,0]-psi_mean[:,c]]
            pl.errorbar(T, psi_mean[:,c], psi_err, fmt='.', color=_color, markersize=7)
            pl.plot(xx, psi_pred[:,c], '-', color=_color, label=_tran[c])
            pl.fill(np.concatenate([xx, xx[::-1]]), np.concatenate([psi_pred95[:,c,0],psi_pred95[:,c,1][::-1]]), 
                    alpha=.15, fc=_color, ec='None')

        pl.xlabel('$t$')
        pl.ylabel('$\psi(t)$')
        pl.ylim(-0.02, 1.02)
        pl.xticks(T, T)
        pl.title(self.gene[qidx])
        pl.legend(loc="best")
        
class DiceFile:
    """docstring for DiceFile"""
    def __init__(self, dice_file):
        dice_data = np.loadtxt(dice_file, skiprows=1, delimiter="\t", dtype="str")
        idx = np.arange(int((dice_data.shape[1]-4) / 3))
        self.gene = dice_data[:,0]
        self.LogLike = dice_data[:,3].astype("float")
        self.GeneCount = dice_data[:,idx*3+4].astype("float")
        
        self.tran, self.tranLen = [], []
        self.IsoRatio, self.CI95 = [], []
        for i in range(len(self.gene)):
            self.tran.append(dice_data[i,1].split(","))
            self.tranLen.append(np.array(dice_data[i,2].split(","), "float"))
            _IsoRatio = np.zeros((len(idx), len(self.tranLen[i])))
            _CI95 = np.zeros((len(idx), len(self.tranLen[i]), 2))
            for j in idx:
                _IsoRatio[j, :] = dice_data[i,j*3+5].split(",")
                _temp = dice_data[i,j*3+6].split(",")
                for k in range(len(_temp)):
                    _CI95[j, k, :] = _temp[k].split(":")
            self.IsoRatio.append(_IsoRatio)
            self.CI95.append(_CI95)


def get_CI(data, percent=0.95):
    """calculate the confidence intervals
    """
    if len(data.shape) == 0:
        data = data.reshape(-1,1)
    RV = np.zeros((data.shape[1],2))
    CI_idx = int(data.shape[0] * (1-percent)/2)
    for k in range(data.shape[1]):
        temp = np.sort(data[:,k])
        RV[k,:] = [temp[-CI_idx], temp[CI_idx]]
    return RV

def GP_K(X, theta, x=None):
    """calculate the covariance matrix
    """
    N = len(X)
    if x is None:
        K = np.zeros((N, N))
        for i in range(N):
            for j in range(N):
                K[i,j] = theta[0] * np.exp(-0.5/theta[1]*(X[i]-X[j])**2)
    else:
        K = np.zeros(N)
        for i in range(N):
            K[i] = theta[0] * np.exp(-0.5/theta[1]*(X[i]-x)**2)
    return K

def GP_predict(Theta, X, Y, Yvar, xx):
    """predict by GP with its parameters
    """
    if len(np.array(xx).shape) == 0:
        xx = np.array([xx])
    yy = np.zeros(xx.shape[0])
    ss = np.zeros(xx.shape[0])

    K0 = GP_K(X, Theta)
    for i in range(K0.shape[0]):
        K0[i,i] += Yvar[i]
    K2 = Theta[0]

    for i in range(xx.shape[0]):
        K1 = GP_K(X, Theta, xx[i])
        yy[i] = np.dot(np.dot(K1, np.linalg.inv(K0)), Y)
        ss[i] = K2 - np.dot(np.dot(K1, np.linalg.inv(K0)), np.transpose(K1))
    return yy, ss

def GP_sample_predict(Theta, X, Yall, xx):
    """predict by GP with its all samples
    """
    if len(np.array(xx).shape) == 0:
        xx = np.array([xx])
    yy = np.zeros((Yall.shape[0], xx.shape[0]))

    K0 = GP_K(X, Theta)
    K2 = Theta[0]

    for i in range(xx.shape[0]):
        idx = np.where(abs(xx[i] - X)<10**(-10))[0]
        if len(idx) == 1:
            yy[:,i] = Yall[:,idx[0]]
            continue
        K1 = GP_K(X, Theta, xx[i])
        for j in range(Yall.shape[0]):
            _y = np.dot(np.dot(K1, np.linalg.inv(K0)), Yall[j,:])
            _v = K2 - np.dot(np.dot(K1, np.linalg.inv(K0)), np.transpose(K1))
            yy[j,i] = np.random.normal(_y, np.sqrt(_v))
    return yy

def YtoPsi(y, axis=0):
    """softmax function: transfer y to psi
    """
    size = np.array(y.shape)
    size[axis] = -1
    return np.exp(y) / (np.exp(y).sum(axis=axis).reshape(size))